"""
Implements embedding edit distance learning as described in the paper

Paaßen, B., Gallicchio, C., Micheli, A., and Hammer, B. (2018).
Tree Edit Distance Learning via Adaptive Symbol Embeddings. Proceedings of
the 35th International Conference on Machine Learning (ICML 2018).
URL: http://proceedings.mlr.press/v80/paassen18a.html

"""
# Copyright (C) 2019-2021
# Benjamin Paaßen
# AG Machine Learning
# Bielefeld University

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from scipy.optimize import minimize
from scipy.spatial.distance import pdist, squareform
from sklearn.base import BaseEstimator, ClassifierMixin
from proto_dist_ml.mglvq import MGLVQ
import edist.sed as sed
import edist.multiprocess as mp

__author__ = 'Benjamin Paaßen'
__copyright__ = 'Copyright (C) 2019-2021, Benjamin Paaßen'
__license__ = 'GPLv3'
__version__ = '1.2.0'
__maintainer__ = 'Benjamin Paaßen'
__email__  = 'bpaassen@techfak.uni-bielefeld.de'

_ERR_CUTOFF = 1E-5
_BFGS_MAX_IT = 100

class BEDL(BaseEstimator, ClassifierMixin):
    """ Implements the embedding edit distance learning (BEDL) scheme for
    metric learning on edit distances.

    In more detail, this learns a median generalized learning vector
    quantization (MGLVQ) classifier on the data and an embedding of the
    input symbols which yields an edit distance that makes classification
    with this classifier easier.

    Attributes
    ----------
    K: int
        The number of prototypes for the MGLVQ classifier.
    T: int
        The number of learning epochs we use at most. Defaults to 5.
    phi: function (default = identity)
        A squashing function to post-process each error term. Defaults to the
        identity.
    phi_grad: function (default = one)
        The gradient function corresponding to phi.
    distance: function (default = sed.sed)
        The edit distance function that shall be learned. Defaults to the
        sequence edit distance sed.sed.
    distance_backtrace: function (default = sed.sed_backtrace_matrix)
        The matrix backtracing function for the distance.
        Defaults to sed.sed_backtrace_matrix. Note that this currently does NOT
        support ADP because ADP returns a different backtracing format.
    _classifier: class proto_dist_ml.MGLVQ
        The learned MGLVQ classifier model.
    _idx: dictionary
        A mapping from alphabet to indices.
    _embedding: array_like
        A len(alphabet) x len(alphabet) - 1 embedding matrix for all symbols in
        the alphabet.
    _delta_obj: class bedl.EmbeddingDelta
        An internal object to make storing of the delta function more
        efficient.
    _delta: function
        The learned delta function.

    """
    def __init__(self, K, T = 5, phi = None, phi_grad = None, distance = None, distance_backtrace = None):
        self.K = K
        self.T = T
        if(phi is None):
            self.phi = lambda mus : mus
            self.phi_grad = lambda mus : np.ones_like(mus)
        else:
            self.phi = phi
        if(distance is None):
            self.distance = sed.sed
            self.distance_backtrace = sed.sed_backtrace_matrix
        else:
            self.distance = distance
            self.distance_backtrace = distance_backtrace

    def fit(self, X, y):
        """ Trains a BEDL model on the given input data.

        In more detail, we iterate the following steps in each training
        epoch:

        1. We re-compute the distance matrix.
        2. We adapt the MGLVQ classifier to the current distance matrix.
        3. We compute the matrix backtraces for the data-to-prototype
           distances.
        4. We optimize the embedding for the current data-to-prototype
           distances.

        For more details, please refer to the ICML 2018 paper.

        Arguments
        ---------
        X: list
            a list of data points, each being either a list or a tree,
            depending on the edit distance that shall be learned.
        y: array_like or list
            an array-like or list-like structure with labels for each
            data point.

        Returns
        -------
        class bedl.BEDL
            self

        """
        # identify the alphabet from the training data
        alphabet = set()
        for i in range(len(X)):
            for t in range(len(X[i])):
                alphabet.add(X[i][t])
        # build an index of the input symbols
        self._idx = create_index(sorted(list(alphabet)))
        # preprocess all input data to have symbol indices as nodes instead
        # of the original symbols
        X = index_data(X, self._idx)
        # initialize the classifier
        self._classifier = MGLVQ(self.K)
        # initialize the embedding
        self._embedding = initialize_embedding(len(self._idx))
        # set up optimizer options
        options = { 'ftol' : _ERR_CUTOFF, 'maxiter' : _BFGS_MAX_IT }
        # set up unique labels
        unique_labels = np.unique(y)
        # keep track of prototype changes
        old_w = None
        # now, start the learning process
        self._loss = []
        for t in range(self.T):
            DeltaObj = EmbeddingDelta(self._embedding)
            # first, compute the current pairwise edit distance matrix
            D = mp.pairwise_distances_symmetric(X, self.distance, DeltaObj.delta)
            # then, train the classifier
            self._classifier.prevent_initialization = t > 0
            self._classifier.fit(D, y)
            self._loss.append(self._classifier._loss[-1])
            # if the prototype locations did not change anymore, stop the
            # optimization early
            if(np.all(old_w == self._classifier._w)):
                break
            # compute the backtracing from datapoints to prototypes
            W = []
            for k in range(len(self._classifier._w)):
                W.append(X[self._classifier._w[k]])
            Ps = mp.pairwise_backtraces(X, W, self.distance_backtrace, DeltaObj.delta)
            # reduce the backtraces to just count the symbol pairings, which
            # speeds up gradient computations later on
            for k in range(len(Ps)):
                for l in range(len(Ps[k])):
                    Ps[k][l] = reduce_backtrace(Ps[k][l][0], X[k], W[l], len(self._embedding))
            # set up the glvq loss and gradient function
            obj = lambda embedding : self._loss_and_grad(embedding, Ps, y, unique_labels)
            # optimize the embedding
            res = minimize(obj, self._embedding.ravel(), method='L-BFGS-B', jac = True, options = options)
            # check optimization result
            if(res.x is None):
                raise ValueError('Optimization returned none and failed with message: %s!' % str(res.message))
            self._loss.append(res.fun)
            self._embedding = res.x.reshape(self._embedding.shape)
            # store current prototypes
            old_w = np.copy(self._classifier._w)
        # store the learned delta function
        self._delta_obj = EmbeddingDelta(self._embedding)
        self._delta_obj._index = self._idx
        self._delta = self._delta_obj.delta_with_indexing
        return self

    def _loss_and_grad(self, embedding, Ps, y, unique_labels):
        """ Computes the GLVQ loss and its gradient with respect to the
        embedding.

        Parameters
        ----------
        embedding: array_like
            the current embedding parameters as a vector.
        Ps: list
            the reduced matrix backtraces between the data and the prototypes.
        y: array_like
            the data labels.
        unique_labels: array_like
            The result of np.unique(y)

        Returns
        -------
        loss: float
            The current GLVQ loss.
        Grad: array_like
            The gradient of the current GLVQ loss with respect to embedding
            in flattened format.

        """
        # reshape embedding back into a matrix if its flattened for optimization
        is_flat = False
        if(len(embedding.shape) == 1):
            is_flat = True
            embedding = embedding.reshape(self._embedding.shape)
        # append zero vector to embedding
        embedding = np.concatenate([embedding, np.zeros((1, embedding.shape[1]))], axis=0)
        # compute the pairwise distances between all embedding elements.
        Delta = squareform(pdist(embedding))
        # compute the datapoint-to-prototype distances based on Ps
        Dp = np.zeros((len(Ps), len(Ps[0])))
        for i in range(len(Ps)):
            for k in range(len(Ps[i])):
                Dp[i, k] = np.sum(Ps[i][k] * Delta)

        # find the closest correct and the closest wrong prototype for all
        # data points
        closest_plus  = np.zeros(len(Ps), dtype=int)
        closest_minus = np.zeros(len(Ps), dtype=int)
        dp = np.zeros(len(Ps))
        dm = np.zeros(len(Ps))
        for l in range(len(unique_labels)):
            # get data points in class l and prototypes in and out of class l.
            inClass_l = np.where(y == unique_labels[l])[0]
            inClass_w_l = np.where(self._classifier._y == unique_labels[l])[0]
            outClass_w_l = np.where(self._classifier._y != unique_labels[l])[0]
            # find the closest prototype in the same class
            closest_plus[inClass_l] = inClass_w_l[np.argmin(Dp[inClass_l, :][:, inClass_w_l], axis=1)]
            dp[inClass_l] = Dp[inClass_l, closest_plus[inClass_l]]
            # find the closest prototype in a different class
            closest_minus[inClass_l] = outClass_w_l[np.argmin(Dp[inClass_l, :][:, outClass_w_l], axis=1)]
            dm[inClass_l] = Dp[inClass_l, closest_minus[inClass_l]]

        # compute the raw loss terms 
        mus = (dp - dm) / (dp + dm + 1E-5)
        # compute the loss
        loss = np.sum(self.phi(mus))
        # compute the gradient for all loss terms
        mus_grad = self.phi_grad(mus) * 2 / np.square(dp + dm + 1E-5)
        # compute the gradients with respect to the embedding
        Grad = np.zeros((embedding.shape[0]-1, embedding.shape[1]))
        for j in range(embedding.shape[0]-1):
            # compute the gradient of the delta function for the current
            # embedding
            delta_grad_j = np.expand_dims(embedding[j, :], 0) - embedding
            nzs = Delta[j, :] > 1E-3
            delta_grad_j[nzs, :] /= np.expand_dims(Delta[j, nzs], 1)
            # multiply that gradient with every P matrix
            for i in range(len(Ps)):
                P_plus = Ps[i][closest_plus[i]]
                Grad[j, :] += mus_grad[i] * dm[i] * np.dot(P_plus[j, :] + P_plus[:, j], delta_grad_j)
                P_minus = Ps[i][closest_minus[i]]
                Grad[j, :] -= mus_grad[i] * dp[i] * np.dot(P_minus[j, :] + P_minus[:, j], delta_grad_j)
        # return the results
        if(is_flat):
            Grad = Grad.flatten()
        return loss, Grad

def create_index(lst):
    """ Creates a map of list elements to indices.

    Parameters
    ----------
    lst: list
        A list.

    Returns
    -------
    idx: dictionary
        A map of list elements to indices.

    """
    idx = {}
    for i in range(len(lst)):
        idx[lst[i]] = i
    return idx


def index_data(Xs, idx):
    """ Indexes all data in the input dataset according to the given index.

    Parameters
    ----------
    Xs: list
        A list of data, each being either a list or a tree in
        node list/adjacency list format.
    idx: dictionary
        A map from symbols to indices.

    Returns
    -------
    Ys: list
        A copy of Xs, where each symbol is replaced by a symbol index.

    """
    Ys = []
    for X in Xs:
        is_tree = isinstance(X, tuple)
        if(is_tree):
            X, adj = X
            if(not isinstance(X, list)):
                raise ValueError('Expected either a list or a tree as data point, but got %s' % str(datum))
        # index the datum
        Y = []
        for x in X:
            if(x not in idx):
                raise ValueError('Symbol not in index: %s' % str(x))
            Y.append(idx[x])
        if(is_tree):
            Ys.append((Y, adj))
        else:
            Ys.append(Y)
    return Ys


def initialize_embedding(size):
    """ Sets up a size-dimensional simplex with side length 1 and size+1
    vertices (i.e. an equilateral hyper-triangle).

    Parameters
    ----------
    size: int
        the dimensionality of the simplex.

    Returns
    -------
    embedding: array_like
        A size x size matrix, where each row contains one vertex of
        the simplex. The origin is the last vertex.

    """
    # set up the initial embedding matrix
    embedding = np.zeros((size, size))
    # set up the rho sequence specifying the vertex positions
    rho = np.zeros(size)
    for j in range(size):
        rho[j] = 1. / (2. * (j+1) * (j+2))
    rho = np.sqrt(rho)
    # set up the embedding itself
    for i in range(size):
        for j in range(i):
            embedding[i, j] = rho[j]
        embedding[i, i] = rho[i] * (i+2)
    return embedding


class EmbeddingDelta:
    """ This class serves as a storage for an embedding to make the
    embedding delta function pickleable.

    Parameters
    ----------
    embedding: array_like
        An embedding matrix.

    Attributes
    ----------
    _Delta: array_like
        The current symbol-to-symbol distance matrix
    """
    def __init__(self, embedding):
        # extend the embedding by a zero vector
        embedding = np.concatenate((embedding, np.zeros((1, embedding.shape[1]))))
        # compute the pairwise distances
        self._Delta = squareform(pdist(embedding))

    def delta(self, x, y):
        """ Computes the distance between two embedding vectors identified
        by their index.

        Parameters
        ----------
        x: int
            the left-hand-side index or None.
        y: int
            the right-hand-side index or None.

        Returns
        -------
        d: float
            The Euclidean distance between the embedding for x and for y.

        """
        if(x is None):
            if(y is None):
                return 0.
            else:
                # return the distance between x and the origin, i.e. the norm
                # of x
                return self._Delta[-1, y]
        if(y is None):
            # return the distance between y and the origin, i.e. the norm of y
            return self._Delta[x, -1]
        return self._Delta[x, y]

    def delta_with_indexing(self, x, y):
        """ Computes the distance between two symbols based on their
        embedding vectors.

        Parameters
        ----------
        x: str
            a symbol.
        y: str
            another symbol.

        Returns
        -------
        d: float
            The Euclidean distance between the embedding for x and for y.
        """
        if(x is None):
            if(y is None):
                return 0.
            else:
                # return the distance between x and the origin, i.e. the norm
                # of x
                return self._Delta[-1, self._index[y]]
        if(y is None):
            # return the distance between y and the origin, i.e. the norm of y
            return self._Delta[self._index[x], -1]
        return self._Delta[self._index[x], self._index[y]]


def reduce_backtrace(P, x, y, size):
    """ Transforms the input backtrace matrix P of size len(x) + 1 x len(y) + 1
    into a reduced matrix Phat of size size + 1 x size + 1 which accumulates the
    probabilities for same symbols being replaced, i.e. Phat[k, l] is the
    sum over all entries P[i, j] where x[i] = k and y[j] = l.

    Parameters
    ----------
    P: array_like
        A len(x) + 1 x len(y) + 1 backtracing matrix between sequence x and y,
        where P[i, j] is the probability of x[i] being replaced with y[j] in a
        co-optimal alignment, P[i, len(y)] is the deletion probability for x[i]
        and P[len(x), j] is the insertion probability for y[j]. The last row
        and column are optional (as in case of DTW).
    X: list
        A list of objects, either sequences or trees.
    Y: list
        A list of objects, either sequences or trees.
    size: int
        The alphabet size.

    Returns
    -------
    Phat: array_like
        A size+1 x size+1 matrix which accumulates the probabilities for same
        symbols being replaced, i.e. Phat[k, l] is the sum over all entries
        P[i, j] where x[i] = k and y[j] = l.

    """
    # extract the node list from trees
    if(isinstance(x, tuple)):
        x = x[0]
    if(isinstance(y, tuple)):
        y = y[0]
    # store where each sequence equals to each symbol
    Loc_x = np.zeros((P.shape[0], size+1), dtype=bool)
    for i in range(len(x)):
        Loc_x[i, x[i]] = True
    if(P.shape[0] > len(x)):
        Loc_x[len(x), size] = True
    Loc_y = np.zeros((P.shape[1], size+1), dtype=bool)
    for j in range(len(y)):
        Loc_y[j, y[j]] = True
    if(P.shape[1] > len(y)):
        Loc_y[len(y), size] = True

    # initialize the accumulation matrix
    Phat = np.zeros((size+1, size+1))
    # accumulate probabilities
    for k in range(size+1):
        for l in range(size+1):
            Phat[k, l] = np.sum(P[Loc_x[:, k], :][:, Loc_y[:, l]])
    return Phat



"""
Implements a sequence edit distance with affine gap costs using ADP.

"""
# Copyright (C) 2019-2021
# Benjamin Paaßen
# AG Machine Learning
# Bielefeld University

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import edist.adp as adp

__author__ = 'Benjamin Paaßen'
__copyright__ = 'Copyright (C) 2019-2021, Benjamin Paaßen'
__license__ = 'GPLv3'
__version__ = '1.2.0'
__maintainer__ = 'Benjamin Paaßen'
__email__  = 'bpaassen@techfak.uni-bielefeld.de'

# first, define the grammar
_grammar = adp.Grammar('A', ['A', 'D', 'I'])
_grammar.append_replacement('A', 'A', 'rep')
_grammar.append_deletion('A', 'D', 'del')
_grammar.append_insertion('A', 'I', 'ins')
_grammar.append_replacement('D', 'A', 'rep')
_grammar.append_deletion('D', 'D', 'skdel')
_grammar.append_insertion('D', 'I', 'ins')
_grammar.append_replacement('I', 'A', 'rep')
_grammar.append_insertion('I', 'I', 'skins')

class AffineAlgebra:
    """ This is a class to efficiently store an algebra for the affine edit
    distance grammar in a pickleable format.

    Attributes
    ----------
    _rep: function (default = Kronecker distance)
        A function for replacement costs, i.e. _rep(x, y) is the cost of
        replacing x with y.
    _gap: function (default = constant function with 1.0)
        A function for deletion/insertion costs, i.e. _gap(x) is the cost of
        deleting/inserting x.
    _gap_cost: float (default = 1.0)
        a constant cost for deletions/insertions.
    _skip: function (default = constant function with 0.5)
        A function for deletion/insertion extension costs, i.e. _skip(x) is the
        cost of skip-deleting/-inserting x.
    _skip_cost: float (default = 0.5)
        a constant cost for deletion/insertion extensions.

    """
    def __init__(self, rep = None, gap = 1., skip = 0.5):
        if(rep is None):
            self._rep = self._kron
        else:
            self._rep = rep
        if(not callable(gap)):
            self._gap = self._gap_const
            self._gap_cost = gap
        else:
            self._gap = gap
        if(not callable(skip)):
            self._skip = self._skip_const
            self._skip_cost = skip
        else:
            self._skip = skip

    def __getitem__(self, key):
        if(key == 'rep'):
            return self._rep
        if(key == 'del' or key == 'ins'):
            return self._gap
        if(key == 'skdel' or key == 'skins'):
            return self._skip
        raise ValueError('Unsupported key: %s' % str(key))

    def __contains__(self, item):
        return item in ['rep', 'del', 'ins', 'skdel', 'skins']

    def _kron(self, x, y):
        if(x == y):
            return 0.
        else:
            return 1.

    def _gap_const(self, x, y):
        return self._gap_cost

    def _skip_const(self, x, y):
        return self._skip_cost

def aed(x, y, rep = None, gap = 1., skip = 0.5):
    """ Computes the affine edit distance using algebraic dynamic programming.

    Parameters
    ----------
    x: list
        A list-like object.
    y: list
        Another list-like object.
    rep: function (default = Kronecker delta)
        A function with two arguments, computing the cost for replacing the
        first with the second OR an AffineAlgebra object, in which case the
        remaining aguments will be ignored. Defaults to the Kronecker distance.
    gap: function or float (default = 1.0)
        A function with two arguments, computing the cost for deleting the
        first or inserting the second OR a number defining a constant cost.
        Defaults to 1.
    skip: function or float (default = 0.5)
        A function with two arguments, computing the cost for deleting the
        first or inserting the second for gap extensions OR a number defining
        a constant cost. Defaults to 0.5.

    Returns
    -------
    d: float
        The affine edit distance between x and y.

    """
    if(isinstance(rep, AffineAlgebra)):
        algebra = rep
    else:
        algebra = AffineAlgebra(rep, gap, skip)
    return adp.edit_distance(x, y, _grammar, algebra)

def aed_backtrace(x, y, rep = None, gap = 1., skip = 0.5):
    """ Computes the backtrace of the affine edit distance using algebraic
    dynamic programming.

    Parameters
    ----------
    x: list
        A list-like object.
    y: list
        Another list-like object.
    rep: function (default = Kronecker delta)
        A function with two arguments, computing the cost for replacing the
        first with the second OR an AffineAlgebra object, in which case the
        remaining aguments will be ignored. Defaults to the Kronecker distance.
    gap: function or float (default = 1.0)
        A function with two arguments, computing the cost for deleting the
        first or inserting the second OR a number defining a constant cost.
        Defaults to 1.
    skip: function or float (default = 0.5)
        A function with two arguments, computing the cost for deleting the
        first or inserting the second for gap extensions OR a number defining
        a constant cost. Defaults to 0.5.

    Returns
    -------
    alignment: class alignment.Alignment
        A co-optimal alignment between x and y according to the affine edit
        distance.

    """
    if(isinstance(rep, AffineAlgebra)):
        algebra = rep
    else:
        algebra = AffineAlgebra(rep, gap, skip)
    return adp.backtrace(x, y, _grammar, algebra)

def aed_backtrace_stochastic(x, y, rep = None, gap = 1., skip = 0.5):
    """ Computes the backtrace of the affine edit distance using algebraic
    dynamic programming stochastically.

    Note that the randomness does _not_ produce a uniform distribution over
    all co-optimal alignments because random choices at the start of the
    alignment process dominate. If you wish to characterize the overall
    distribution accurately, use aed_backtrace_matrix instead.

    Parameters
    ----------
    x: list
        A list-like object.
    y: list
        Another list-like object.
    rep: function (default = Kronecker delta)
        A function with two arguments, computing the cost for replacing the
        first with the second OR an AffineAlgebra object, in which case the
        remaining aguments will be ignored. Defaults to the Kronecker distance.
    gap: function or float (default = 1.0)
        A function with two arguments, computing the cost for deleting the
        first or inserting the second OR a number defining a constant cost.
        Defaults to 1.
    skip: function or float (default = 0.5)
        A function with two arguments, computing the cost for deleting the
        first or inserting the second for gap extensions OR a number defining
        a constant cost. Defaults to 0.5.

    Returns
    -------
    alignment: class alignment.Alignment
        A co-optimal alignment between x and y according to the affine edit
        distance.

    """
    if(isinstance(rep, AffineAlgebra)):
        algebra = rep
    else:
        algebra = AffineAlgebra(rep, gap, skip)
    return adp.backtrace_stochastic(x, y, _grammar, algebra)

def aed_backtrace_matrix(x, y, rep = None, gap = 1., skip = 0.5):
    """ Computes the backtrace matrix P of the affine edit distance using
    algebraic dynamic programming.

    In particular, P[i, j] contains the probability of node i being replaced
    with node j in a co-optimal alignment. The last two columns contain
    deletion and deletion-extension probabilities, the last two rows contains
    insertion and insertion-extension probabilities.

    Parameters
    ----------
    x: list
        A list-like object.
    y: list
        Another list-like object.
    rep: function (default = Kronecker distance)
        A function with two arguments, computing the cost for replacing the
        first with the second OR an AffineAlgebra object, in which case the
        remaining aguments will be ignored. Defaults to the Kronecker distance.
    gap: function or float (default = 1.0)
        A function with two arguments, computing the cost for deleting the
        first or inserting the second OR a number defining a constant cost.
        Defaults to 1.
    skip: function or float (default = 0.5)
        A function with two arguments, computing the cost for deleting the
        first or inserting the second for gap extensions OR a number defining
        a constant cost. Defaults to 0.5.

    Returns
    -------
    P: array_like
        A len(x) + 2 x len(y) + 2 matrix where P[i, j] contains the probability
        of node i being replaced with node j in a co-optimal alignment. The last
        two columns contain deletion and deletion-extension probabilities, the
        last two rows contains insertion and insertion-extension probabilities.
    k: int
        The number of co-optimal alignments.

    """
    if(isinstance(rep, AffineAlgebra)):
        algebra = rep
    else:
        algebra = AffineAlgebra(rep, gap, skip)
    P_rep, P_del, P_ins, k = adp.backtrace_matrix(x, y, _grammar, algebra)
    # re-format the matrices
    P = np.zeros((len(x) + 2, len(y) + 2))
    P[:len(x), :][:, :len(y)] = P_rep[0, :, :]
    P[:len(x), len(y):] = P_del.T
    P[len(x):, :len(y)] = P_ins
    # return the result
    return P, k



#!python
#cython: language_level=3
"""
Implements the tree edit distance of Zhang and Shasha (1989) and its
backtracing in cython.

"""
# Copyright (C) 2019-2021
# Benjamin Paaßen
# AG Machine Learning
# Bielefeld University

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import heapq
import numpy as np
from cython.parallel import prange
from libc.math cimport sqrt
from cpython cimport bool
cimport cython
from edist.alignment import Alignment

__author__ = 'Benjamin Paaßen'
__copyright__ = 'Copyright (C) 2019-2021, Benjamin Paaßen'
__license__ = 'GPLv3'
__version__ = '1.2.0'
__maintainer__ = 'Benjamin Paaßen'
__email__  = 'bpaassen@techfak.uni-bielefeld.de'

###################################
# Edit Distance with Custom Delta #
###################################

def ted(x_nodes, x_adj, y_nodes = None, y_adj = None, delta = None):
    """ Computes the tree edit distance between the trees x and y, each
    described by a list of nodes and an adjacency list adj, where adj[i]
    is a list of indices pointing to children of node i.

    Note that we assume a proper depth-first-search order of adj, i.e. for
    every node i, the following indices are all part of the subtree rooted at
    i until we hit the index of i's right sibling or the end of the tree.

    Parameters
    ----------
    x_nodes: list or tuple
        a list of nodes for tree x OR a tuple of the form (x_nodes, x_adj).
    x_adj: list or tuple
        an adjacency list for tree x OR a tuple of the form (y_nodes, y_adj).
    y_nodes: list (default = x_adj[0])
        a list of nodes for tree y.
    y_adj: list (default = x_adj[1])
        an adjacency list for tree y.
    delta: function (default = None)
        a function that takes two nodes as inputs and returns their pairwise
        distance, where delta(x, None) should be the cost of deleting x and
        delta(None, y) should be the cost of inserting y. If undefined, this
        method calls standard_ted instead.

    Returns
    -------
    d: float
        the tree edit distance between x and y according to delta.

    """
    if(delta is None):
        return float(standard_ted(x_nodes, x_adj, y_nodes, y_adj))

    if(isinstance(x_nodes, tuple)):
        x_nodes, x_adj, y_nodes, y_adj = extract_from_tuple_input(x_nodes, x_adj)
    # the number of nodes in both trees
    cdef int m = len(x_nodes)
    cdef int n = len(y_nodes)
    # if either tree is empty, we can only delete/insert all nodes in the
    # non-empty tree.
    cdef double d = 0
    if(m == 0):
        for j in range(n):
            d += delta(None, y_nodes[j])
        return d
    if(n == 0):
        for i in range(m):
            d += delta(x_nodes[i], None)
        return d
    # otherwise, compute the actual tree edit distance
    _, _, _, _, _, _, D_tree = _ted(x_nodes, x_adj, y_nodes, y_adj, delta)
    return D_tree[0,0]

def _ted(x_nodes, x_adj, y_nodes = None, y_adj = None, delta = None):
    """ Internal function; call ted instead. """
    if(isinstance(x_nodes, tuple)):
        x_nodes, x_adj, y_nodes, y_adj = extract_from_tuple_input(x_nodes, x_adj)

    # the number of nodes in both trees
    cdef int m = len(x_nodes)
    cdef int n = len(y_nodes)
    # An array to store all edit costs for replacements, deletions, and
    # insertions
    Delta = np.zeros((m+1, n+1))
    cdef double[:,:] Delta_view = Delta
    # First, compute all pairwise replacement costs
    cdef int i
    cdef int j
    for i in range(m):
        for j in range(n):
            Delta_view[i,j] = delta(x_nodes[i], y_nodes[j])

    # Then, compute the deletion and insertion costs
    for i in range(m):
        Delta_view[i,n] = delta(x_nodes[i], None)
    for j in range(n):
        Delta_view[m,j] = delta(None, y_nodes[j])

    # Compute the keyroots and outermost right leaves for both trees.
    x_orl = outermost_right_leaves(x_adj)
    x_kr  = keyroots(x_orl)
    y_orl = outermost_right_leaves(y_adj)
    y_kr  = keyroots(y_orl)

    # Finally, compute the actual tree edit distance
    D_forest = np.zeros((m+1,n+1))
    D_tree = np.zeros((m,n))
    _ted_c(x_orl, x_kr, y_orl, y_kr, Delta, D_forest, D_tree)
    return x_orl, x_kr, y_orl, y_kr, Delta, D_forest, D_tree

def extract_from_tuple_input(x, y):
    """ Assumes that both x and y are tuples and unpacks those tuples.

    Parameters:
    -----------
    x: tuple
        a tuple
    y: tuple
        another tuple

    Returns
    -------
    x_nodes: list
        x[0]
    x_adj: list
        x[1]
    y_nodes: list
        y[0]
    y_adj: list
        y[1]

    Raises
    ------
    ValueError
        if the input is not tuple-shaped.

    """
    if(len(x) != 2):
        raise ValueError('If the first input is a tuple, it needs to contain exactly two elements, a node list and an adjacency list')
    if(not isinstance(y, tuple)):
        raise ValueError('If the first input is a tuple, the second input needs to be a tuple as well!')
    if(len(y) != 2):
        raise ValueError('If the second input is a tuple, it needs to contain exactly two elements, a node list and an adjacency list')
    x_nodes = x[0]
    x_adj   = x[1]
    y_nodes = y[0]
    y_adj   = y[1]
    return x_nodes, x_adj, y_nodes, y_adj

def outermost_right_leaves(list adj):
    """ Computes the outermost right leaves of a tree based on its adjacency
    list. The outermost right leaf of a tree is defined as recursively
    accessing the right-most child of a node until we hit a leaf.

    Note that we assume a proper depth-first-search order of adj, i.e. for
    every node i, the following indices are all part of the subtree rooted at
    i until we hit the index of i's right sibling or the end of the tree.

    Parameters
    ----------
    adj: list
        An adjacency list representation of the tree, i.e. an array such that
        for every i, adj[i] is the list of child indices for node i.

    Returns
    -------
    orl: int array
        An array containing the outermost right leaf index for every node
        in the tree.

    """
    # the number of nodes in the tree
    cdef int m = len(adj)
    # the array into which we will write the outermost right leaves for each
    # node
    orl = np.full(m, -1, dtype=int)
    cdef long[:] orl_view = orl
    # a temporary variable for the current outermost right leaf
    cdef int r
    # iterate over all nodes to retrieve the respective outermost right leave
    cdef int i
    for i in range(m):
        # keep searching until we hit a node for which the outermost right
        # leaf is already defined or until we hit a leaf
        r = i
        while(True):
            # if r has no children, r is the outermost right leaf for i
            if(not adj[r]):
                orl_view[i] = r
                break
            # if the outermost right leaf for r is defined, that is also the
            # outermost right leaf for i
            if(orl_view[r] >= 0):
                orl_view[i] = orl_view[r]
                break
            # otherwise, continue searching
            r = adj[r][-1]
    return orl


def keyroots(long[:] orl):
    """ Computes the keyroots of a tree based on its outermost right leaf
    array. The keyroot for a node i is defined as the lowest k, such that
    orl[i] = orl[k].

    Parameters
    ----------
    orl: array_like
        An outermost right leaf array as computed by the
        outermost_right_leaves function above.

    Returns
    -------
    keyroots: int array
        An array of keyroots in descending order.

    """
    # the number of nodes in the tree
    cdef int m = len(orl)
    # a temporary array to store the keyroots for each outermost right leaf
    kr = np.full(m, -1, dtype=int)
    cdef long[:] kr_view = kr
    # a variable to count the number of keyroots
    cdef int K = 0
    # iterate over all nodes
    cdef int i
    cdef int r
    for i in range(m):
        # check if a keyroot has been found for the current outermost
        # right leaf already
        r = orl[i]
        if(kr_view[r] < 0):
            # if not, store the new keyroot and increment the number of
            # keyroots
            kr_view[r] = i
            K += 1
    # in a next iteration, generate a new array which only contains
    # the defined keyroots
    keyroots = np.zeros(K, dtype=int)
    cdef long[:] keyroots_view = keyroots
    # insert and sort via insertionsort
    # counting index for the current keyroot
    cdef int k
    # insertion index for the current keyroot
    cdef int j
    # index for the current node
    i = 0
    for k in range(K):
        # iterate until the next keyroot is found
        while(kr_view[i] < 0):
            i += 1
        # decrement j until we found the location where the new keyroot should
        # be inserted
        j = k
        while(j > 0 and keyroots_view[j-1] < kr_view[i]):
            keyroots_view[j] = keyroots_view[j-1]
            j -= 1
        # insert the new keyroot
        keyroots_view[j] = kr_view[i]
        # increment i once more
        i += 1
    # return the resulting array
    return keyroots


@cython.boundscheck(False)
cdef void _ted_c(const long[:] x_orl, const long[:] x_kr, const long[:] y_orl, const long[:] y_kr, const double[:,:] Delta, double[:,:] D, double[:,:] D_tree) nogil:
    """ This method is internal and performs the actual tree edit distance
    computation for trees x and y in pure C.

    For details on the algorithm, please refer to the following tutorial:
    https://arxiv.org/abs/1805.06869

    Let m and n be the size of tree x and y respectively.

    Parameters
    ----------
    x_orl: long array
        the outermost right leaves for tree x (int array of length m).
    x_kr: long array
        the keyroots for tree x in descending order (int array).
    y_orl: long array
        the outermost right leaves for tree y (int array of length n).
    y_kr: long array
        the keyroots for tree y in descending order (int array).
    Delta: double matrix
        an (m+1) x (n+1) matrix, where Delta[i,j] for i < m, j < n is the
        cost of replacing x[i] with y[j], where Delta[i,n] is the cost of
        deleting x[i], and where Delta[m,j] is the cost of inserting y[j].
    D: double matrix
        an empty (m+1) x (n+1) matrix used for temporary computations.
    D_tree: double matrix
        an empty m x n matrix. After this method has run, D_tree[i,j] will
        be the tree edit distance between the subtree rooted at i and the
        subtree rooted at j.

    """
    # the number of nodes in both trees
    cdef int m = len(x_orl)
    cdef int n = len(y_orl)
    # the number of keyroots in both trees
    cdef int K = len(x_kr)
    cdef int L = len(y_kr)

    # set up iteration variables
    # for the keyroots
    cdef int k
    cdef int l
    # for the nodes in the subtrees rooted at the keyroots
    cdef long i
    cdef long j
    # and temporary variables for the keyroots and the outermost right leaves
    cdef long i_0
    cdef long j_0
    cdef long i_max
    cdef long j_max

    # iterate over all pairwise combinations of keyroots
    for k in range(K):
        for l in range(L):
            # We consider now the subtree rooted at x_kr[k] versus the subtree
            # rooted at y_kr[l]. The forest edit distances between these
            # subtrees correspond exactly to the matrix block
            # D[x_kr[k]:x_orl[x_kr[k]]+1, y_kr[l]:y_orl[y_kr[l]]+1],
            # which we compute now.
            i_0 = x_kr[k]
            j_0 = y_kr[l]
            i_max = x_orl[i_0] + 1
            j_max = y_orl[j_0] + 1
            # first, initialize the last entry for the current subtree
            # computation
            D[i_max, j_max] = 0.
            # then, initialize the last column
            for i in range(i_max-1, i_0-1, -1):
                D[i, j_max] = Delta[i, n] + D[i+1, j_max]
            # then, initialize the last row
            for j in range(j_max-1, j_0-1, -1):
                D[i_max, j] = Delta[m, j] + D[i_max, j+1]
            # finally, compute the remaining forest edit distances
            for i in range(i_max-1, i_0-1, -1):
                for j in range(j_max-1, j_0-1, -1):
                    if(x_orl[i] == i_max-1 and y_orl[j] == j_max-1):
                        # if we consider a complete subtree, the forest edit
                        # distance D[i,j] is equal to the tree edit distance
                        # at that position and we can compute it via the
                        # standard edit distance recurrence
                        D[i,j] = min3(Delta[i,j] + D[i+1,j+1], # replacement
                                      Delta[i,n] + D[i+1,j], # deletion
                                      Delta[m,j] + D[i,j+1] # insertion
                                 )
                        # store the newly computed tree edit distance as well
                        D_tree[i,j] = D[i,j]
                    else:
                        # if we do _not_ consider a complete subtree, replacements
                        # are only possible between entire subtrees, which we have
                        # to consider in recurrence
                        D[i,j] = min3(D_tree[i,j] + D[x_orl[i]+1,y_orl[j]+1], # tree replacement
                                      Delta[i,n] + D[i+1,j], # deletion
                                      Delta[m,j] + D[i,j+1] # insertion
                                 )


cdef double min3(double a, double b, double c) nogil:
    """ Computes the minimum of three numbers.

    Parameters
    ----------
    a: double
        a number
    b: double
        another number
    c: double
        yet another number

    Returns
    -------
    min3: double
        min({a, b, c})

    """
    if(a < b):
        if(a < c):
            return a
        else:
            return c
    else:
        if(b < c):
            return b
        else:
            return c

#########################
# Backtracing Functions #
#########################

cdef double _BACKTRACE_TOL = 1E-5

def ted_backtrace(x_nodes, x_adj, y_nodes = None, y_adj = None, delta = None):
    """ Computes the tree edit distance between the trees x and y, each
    described by a list of nodes and an adjacency list adj, where adj[i]
    is a list of indices pointing to children of node i. This function
    returns an alignment representation of the distance.

    Note that we assume a proper depth-first-search order of adj, i.e. for
    every node i, the following indices are all part of the subtree rooted at
    i until we hit the index of i's right sibling or the end of the tree.

    Parameters
    ----------
    x_nodes: list or tuple
        a list of nodes for tree x OR a tuple of the form (x_nodes, x_adj).
    x_adj: list or tuple
        an adjacency list for tree x OR a tuple of the form (y_nodes, y_adj).
    y_nodes: list (default = x_adj[0])
        a list of nodes for tree y.
    y_adj: list (default = x_adj[1])
        an adjacency list for tree y.
    delta: function (default = None)
        a function that takes two nodes as inputs and returns their pairwise
        distance, where delta(x, None) should be the cost of deleting x and
        delta(None, y) should be the cost of inserting y. If undefined, this
        method calls standard_ted instead.

    Returns
    -------
    alignment: class alignment.Alignment
        A co-optimal alignment between x and y according to the tree edit
        distance.

    """
    if(delta is None):
        return standard_ted_backtrace(x_nodes, x_adj, y_nodes, y_adj)
    x_orl, x_kr, y_orl, y_kr, Delta, D, D_tree = _ted(x_nodes, x_adj, y_nodes, y_adj, delta)

    # initialize the alignment
    ali = Alignment()

    # start the recursive backtrace computation
    _ted_backtrace(x_orl, y_orl, Delta, D, D_tree, ali, 0, 0)
    return ali

def _ted_backtrace(const long[:] x_orl, const long[:] y_orl, const double[:,:] Delta, double[:,:] D, const double[:,:] D_tree, ali, int k, int l):
    """ Internal function; call ted_backtrace instead.

        Performs the backtracing for the subtree rooted at k in x versus the
        subtree rooted at l in y.
    """
    # recompute the dynamic programming matrix for the current subtree
    # combination
    cdef int m = len(x_orl)
    cdef int n = len(y_orl)
    cdef int i_max = x_orl[k] + 1
    cdef int j_max = y_orl[l] + 1
    cdef int i
    cdef int j
    if(k > 0 or l > 0):
        # note that D[i_max, j_max] is already correctly initialized
        # initialize the last column
        for i in range(i_max-1, k-1, -1):
            D[i, j_max] = Delta[i, n] + D[i+1, j_max]
        # then, initialize the last row
        for j in range(j_max-1, l-1, -1):
            D[i_max, j] = Delta[m, j] + D[i_max, j+1]
        # finally, compute the remaining forest edit distances
        for i in range(i_max-1, k-1, -1):
            for j in range(j_max-1, l-1, -1):
                if(x_orl[i] == x_orl[k] and y_orl[j] == y_orl[l]):
                    # if we consider a complete subtree, we can re-use
                    # the tree edit distance values we computed in the
                    # forward pass
                    D[i,j] = D_tree[i,j] + D[i_max, j_max]
                else:
                    # if we do _not_ consider a complete subtree,
                    # replacements are only possible between entire
                    # subtrees, which we have to consider in
                    # recurrence
                    D[i,j] = min3(D_tree[i,j] + D[x_orl[i]+1,y_orl[j]+1], # tree replacement
                                  Delta[i,n] + D[i+1,j], # deletion
                                  Delta[m,j] + D[i,j+1]  # insertion
                             )
    # now, start the backtracing for the current subtree combination
    i = k
    j = l
    while(i < i_max and j < j_max):
        # check whether a deletion is co-optimal
        if(D[i, j] + _BACKTRACE_TOL > Delta[i, n] + D[i+1, j]):
            # if so, append a deletion operation, increment i, and continue
            ali.append_tuple(i, -1)
            i += 1
            continue
        # check whether an insertion is co-optimal
        if(D[i, j] + _BACKTRACE_TOL > Delta[m, j] + D[i, j+1]):
            # if so, append an insertion operation, increment j, and continue
            ali.append_tuple(-1, j)
            j += 1
            continue
        # check wehther replacement is co-optimal. In this case, we need to
        # consider two cases
        if(x_orl[i] == x_orl[k] and y_orl[j] == y_orl[l]):
            # If we are at the root of postfix-subtrees for subtree k and l,
            # we consider the standard replacement case
            if(D[i,j] + _BACKTRACE_TOL > Delta[i,j] + D[i+1,j+1]):
                # append a replacement operation, increment i and j, and
                # continue
                ali.append_tuple(i, j)
                i += 1
                j += 1
                continue
        else:
            if(D[i, j] + _BACKTRACE_TOL > D_tree[i,j] + D[x_orl[i]+1,y_orl[j]+1]):
                # Otherwise, we consider the case where we replace the entire
                # subtree rooted at i with the entire subtree rooted at j.
                # For this case, we call the backtracing recursively
                _ted_backtrace(x_orl, y_orl, Delta, D, D_tree, ali, i, j)
                i = x_orl[i]+1
                j = y_orl[j]+1
                continue
        # if we got here, nothing is co-optimal, which is an error
        raise ValueError('Internal error: No option is co-optimal.')
    # delete and insert any remaining nodes
    while(i < i_max):
        ali.append_tuple(i, -1)
        i += 1
    while(j < j_max):
        ali.append_tuple(-1, j)
        j += 1

def ted_backtrace_matrix(x_nodes, x_adj, y_nodes = None, y_adj = None, delta = None):
    """ Computes a matrix P where entry P[i, j] represents how often node
    i in tree x was aligned with node j in tree y in co-optimal alignments
    according to the tree edit distance.

    Note that we assume a proper depth-first-search order of adj, i.e. for
    every node i, the following indices are all part of the subtree rooted at
    i until we hit the index of i's right sibling or the end of the tree.

    Parameters
    ----------
    x_nodes: list or tuple
        a list of nodes for tree x OR a tuple of the form (x_nodes, x_adj).
    x_adj: list or tuple
        an adjacency list for tree x OR a tuple of the form (y_nodes, y_adj).
    y_nodes: list (default = x_adj[0])
        a list of nodes for tree y.
    y_adj: list (default = x_adj[1])
        an adjacency list for tree y.
    delta: function (default = None)
        a function that takes two nodes as inputs and returns their pairwise
        distance, where delta(x, None) should be the cost of deleting x and
        delta(None, y) should be the cost of inserting y. If undefined, this
        method calls standard_ted instead.

    Returns
    -------
    P: array_like
        a matrix, where entry P[i, j] specifies the fraction of co-optimal
        alignments in which node x[i] has been aligned with node y[j].
        P[i, n] contains the fraction of deletions of node x[i] and P[m, j]
        the fraction of insertions of node y[j].
    K: array_like
        a matrix that contains the counts for all co-optimal alignments in
        which node x[i] has been aligned with node y[j].
    k: int
        the number of co-optimal alignments overall, such that P = K / k.

    """
    if(delta is None):
        raise ValueError('Not yet supported!')
        # return standard_ted_backtrace_matrix(x_nodes, x_adj, y_nodes, y_adj)

    if(isinstance(x_nodes, tuple)):
        x_nodes, x_adj, y_nodes, y_adj = extract_from_tuple_input(x_nodes, x_adj)

    m = len(x_nodes)
    n = len(y_nodes)
    # compute tree edit distance first
    x_orl, x_kr, y_orl, y_kr, Delta, D, D_tree = _ted(x_nodes, x_adj, y_nodes, y_adj, delta)

    # set up a dictionary to sparsely store the counting matrices for all subtrees
    Ks = {}
    # set up a matrix to store the number of co-optimal alignments for all subtrees
    Kappa = np.zeros((m, n), dtype=int)

    # start the recursive backtrace computation
    _ted_backtrace_matrix(x_orl, x_kr, y_orl, y_kr, Delta, D, D_tree, Ks, Kappa, 0, 0)

    # extract results
    K = Ks[(0,0)]
    k = Kappa[0, 0]
    # construct P
    P = np.zeros((m+1, n+1))
    P[:m, :][:, :n] = K
    P[:m, n] = k - np.sum(K, axis=1)
    P[m, :n] = k - np.sum(K, axis=0)
    P /= k
    # return results
    return P, K, k

def _ted_backtrace_matrix(const long[:] x_orl, const long[:] x_kr, const long[:] y_orl, const long[:] y_kr, const double[:,:] Delta, double[:,:] D, const double[:,:] D_tree, Ks, long[:,:] Kappa, int k, int l):
    """ Internal function; call ted_backtrace_matrix instead.

        Performs the backtracing for the subtree rooted at k in x versus the
        subtree rooted at l in y.
    """

    # get the sizes of the current subtrees
    cdef int m_k = x_orl[k] - k + 1
    cdef int n_l = y_orl[l] - l + 1
    if(m_k == 1 and n_l == 1):
        # if this is a pair of leaves, the handling is trivial
        Kappa[k, l] = 1
        Ks[(k,l)] = np.array([[1]], dtype=int)
        return

    cdef int m = len(x_orl)
    cdef int n = len(y_orl)

    # compute the forward matrix Alpha, which contains the number of
    # co-optimal alignment paths from cell [k, l, k, l] to cell [k, l, i, j]
    Alpha = np.zeros((m_k+1, n_l+1), dtype=int)
    cdef long[:, :] Alpha_view = Alpha
    # we start with a single path
    Alpha_view[0, 0] = 1
    cdef int i
    cdef int j
    cdef double[:,:] D_kl
    if(k == 0 and l == 0):
        D_kl = D
    else:
        D_kl = np.zeros((m_k+1, n_l+1))

    if(k > 0 or l > 0):
        # if we are not considering the entire tree, we always use a
        # replacement as first action to avoid duplicates with other options
        Alpha_view[1, 1] = 1
        # re-compute the edit costs for the current subtree combination
        D_kl[m_k][n_l] = D[x_orl[k], y_orl[l]]
        D_kl[0][0]     = D_tree[k][l] + D[x_orl[k], y_orl[l]]
        for i in range(m_k-1, 0, -1):
            D_kl[i, n_l] = Delta[k+i, n] + D_kl[i+1, n_l]
        # then, initialize the last row
        for j in range(n_l-1, 0, -1):
            D_kl[m_k, j] = Delta[m, l+j] + D_kl[m_k, j+1]
        # finally, compute the remaining forest edit distances
        for i in range(m_k-1, 0, -1):
            for j in range(n_l-1, 0, -1):
                if(x_orl[k+i] == x_orl[k] and y_orl[l+j] == y_orl[l]):
                    # if we consider a complete subtree, we can re-use
                    # the tree edit distance values we computed in the
                    # forward pass
                    D_kl[i,j] = D_tree[k+i,l+j] + D_kl[m_k][n_l]
                else:
                    # if we do _not_ consider a complete subtree,
                    # replacements are only possible between entire
                    # subtrees, which we have to consider in
                    # recurrence
                    D_kl[i,j] = min3(D_tree[k+i,l+j] + D_kl[x_orl[k+i]-k+1,y_orl[l+j]-l+1], # tree replacement
                                  Delta[k+i,n] + D_kl[i+1,j], # deletion
                                  Delta[m,l+j] + D_kl[i,j+1]  # insertion
                             )
        # start our processing queue at 1,1
        q = [(1, 1)]
    else:
        # start our processing queue at 0,0
        q = [(0, 0)]
    # build a set which stores the already visited cells
    visited = set()
    # search along co-optimal paths until the processing queue is empty
    cdef int found_coopt
    cdef int itar
    cdef int jtar
    while(q):
        (i, j) = heapq.heappop(q)
        if((i, j) in visited):
            continue
        visited.add((i, j))
        num_coopts = Alpha_view[i, j]
        if(i == m_k):
            if(j == n_l):
                continue
            # if we are at the end of the first subtree, we can only insert
            if(D_kl[i, j] + _BACKTRACE_TOL > Delta[m,l+j] + D_kl[i,j+1]):
                Alpha_view[i, j+1] += num_coopts
                heapq.heappush(q, (i, j+1))
            else:
                raise ValueError('Internal error: No option is co-optimal.')
            continue
        if(j == n_l):
            # if we are at the end of the second subtree, we can only delete
            if(D_kl[i, j] + _BACKTRACE_TOL > Delta[k+i,n] + D_kl[i+1, j]):
                Alpha_view[i+1, j] += num_coopts
                heapq.heappush(q, (i+1, j))
            else:
                raise ValueError('Internal error: No option is co-optimal.')
            continue

        found_coopt = False
        if(D_kl[i, j] + _BACKTRACE_TOL > Delta[k+i,n] + D_kl[i+1, j]):
            # deletion is co-optimal
            Alpha_view[i+1, j] += num_coopts
            heapq.heappush(q, (i+1, j))
            found_coopt = True
        if(D_kl[i, j] + _BACKTRACE_TOL > Delta[m,l+j] + D_kl[i,j+1]):
            # insertion is co-optimal
            Alpha_view[i, j+1] += num_coopts
            heapq.heappush(q, (i, j+1))
            found_coopt = True
        if(Delta[k+i, l+j] + _BACKTRACE_TOL > Delta[k+i,n] + Delta[m,l+j]):
            # if replacement is as expensive as deletion and insertion, we need
            # to prevent counting replacements because we would overcount
            # otherwise
            continue
        if(x_orl[k+i] == x_orl[k] and y_orl[l+j] == y_orl[l]):
            # If we are at the root of postfix-subtrees for subtree k and l,
            # we consider the standard replacement case
            if(D_kl[i,j] + _BACKTRACE_TOL > Delta[k+i,l+j] + D_kl[i+1,j+1]):
                # replacement is co-optimal
                Alpha_view[i+1, j+1] += num_coopts
                heapq.heappush(q, (i+1, j+1))
                found_coopt = True
        else:
            itar = x_orl[k+i]-k+1
            jtar = y_orl[l+j]-l+1
            if(D_kl[i, j] + _BACKTRACE_TOL > D_tree[k+i,l+j] + D_kl[itar, jtar]):
                # Otherwise, we consider the case where we replace the entire
                # subtree rooted at i with the entire subtree rooted at j.
                # For this case, we call the backtracing recursively
                if((k+i, l+j) not in Ks):
                    _ted_backtrace_matrix(x_orl, x_kr, y_orl, y_kr, Delta, D, D_tree, Ks, Kappa, k+i, l+j)
                # then, we can use the number of paths during recursion,
                # multiplied with the number of coopts we have accumulated so
                # far
                Alpha_view[itar][jtar] += num_coopts * Kappa[k+i,l+j]
                heapq.heappush(q, (itar, jtar))
                found_coopt = True
        if(not found_coopt):
            raise ValueError('Internal error: No option is co-optimal.')

    # store the number of co-optimals for this subtree
    Kappa[k, l] = Alpha_view[m_k, n_l]

    # next, we compute the backward counting matrix, Beta, which contains the
    # number of co-optimal alignment paths from cell [k, l, i, j] to cell
    # [k, l, m_k, n_l]
    Beta = np.zeros((m_k+1, n_l+1), dtype=int)
    cdef long[:, :] Beta_view = Beta
    Beta_view[m_k, n_l] = 1
    # add (0,0) to the visited set to ensure consistency because we started
    # at (1, 1) during forward computation
    visited.add((0,0))
    # iterate in downward lexigraphic order over the visited cells
    for (i, j) in sorted(visited, reverse = True):
        if(i == m_k):
            if(j == n_l):
                continue
            # if we are at the end of the first subtree, we can only insert
            if(D_kl[i, j] + _BACKTRACE_TOL > Delta[m,l+j] + D_kl[i,j+1]):
                Beta_view[i, j] += Beta_view[i, j+1]
            else:
                raise ValueError('Internal error: No option is co-optimal.')
            continue
        if(j == n_l):
            # if we are at the end of the second subtree, we can only delete
            if(D_kl[i, j] + _BACKTRACE_TOL > Delta[k+i,n] + D_kl[i+1, j]):
                Beta_view[i, j] += Beta_view[i+1,j]
            else:
                raise ValueError('Internal error: No option is co-optimal.')
            continue

        found_coopt = False
        if(D_kl[i, j] + _BACKTRACE_TOL > Delta[k+i,n] + D_kl[i+1, j]):
            # deletion is co-optimal
            Beta_view[i, j] += Beta_view[i+1, j]
            found_coopt = True
        if(D_kl[i, j] + _BACKTRACE_TOL > Delta[m,l+j] + D_kl[i,j+1]):
            # insertion is co-optimal
            Beta_view[i, j] += Beta_view[i, j+1]
            found_coopt = True
        if(Delta[k+i, l+j] + _BACKTRACE_TOL > Delta[k+i,n] + Delta[m,l+j]):
            # if replacement is as expensive as deletion and insertion, we need
            # to prevent counting replacements because we would overcount
            # otherwise
            continue
        if(x_orl[k+i] == x_orl[k] and y_orl[l+j] == y_orl[l]):
            # If we are at the root of postfix-subtrees for subtree k and l,
            # we consider the standard replacement case
            if(D_kl[i,j] + _BACKTRACE_TOL > Delta[k+i,l+j] + D_kl[i+1,j+1]):
                # replacement is co-optimal
                Beta_view[i, j] += Beta_view[i+1, j+1]
                found_coopt = True
        else:
            itar = x_orl[k+i]-k+1
            jtar = y_orl[l+j]-l+1
            if(D_kl[i, j] + _BACKTRACE_TOL > D_tree[k+i,l+j] + D_kl[itar, jtar]):
                # if we replace an entire subtree, we need to consider the
                # number of co-optimal alignments between those, which is
                # listed in Kappa
                Beta_view[i][j] += Beta_view[itar][jtar] * Kappa[k+i,l+j]
                found_coopt = True
        if(not found_coopt):
            raise ValueError('Internal error: No option is co-optimal.')

    if(Alpha_view[m_k, n_l] != Beta_view[0, 0]):
        raise ValueError('Internal error: Alignment count in Alpha and Beta matrix did not agree; got %d versus %d' % (Alpha_view[m_k, n_l], Beta_view[0, 0]))

    # initialize the counting matrix for the current subtree
    K = np.zeros((m_k, n_l), dtype=int)
    cdef long[:,:] K_view = K
    cdef long[:,:] K_ij
    cdef int i2
    cdef int j2
    # compute content of K
    for (i, j) in visited:
        if(i == m_k or j == n_l):
            continue
        if((x_orl[k+i] == x_orl[k] and y_orl[l+j] == y_orl[l]) or
           (Delta[k+i, l+j] + _BACKTRACE_TOL > Delta[k+i,n] + Delta[m,l+j])):
            # If we are at the root of postfix-subtrees for subtree k and l
            # _or_ if replacements are as expensive as deletions plus insertions,
            # we count replacements directly
            if(D_kl[i,j] + _BACKTRACE_TOL > Delta[k+i,l+j] + D_kl[i+1,j+1]):
                K_view[i, j] += Alpha_view[i,j] * Beta_view[i+1,j+1]
        else:
            itar = x_orl[k+i]-k+1
            jtar = y_orl[l+j]-l+1
            if(D_kl[i, j] + _BACKTRACE_TOL > D_tree[k+i,l+j] + D_kl[itar, jtar]):
                # if we replace an entire subtree, we need to consider the
                # number of co-optimal alignments between those, which is
                # listed in Kappa
                num_coopts = Alpha_view[i,j] * Beta_view[itar,jtar]
                K_ij = Ks[(k+i, l+j)]
                for i2 in range(i, itar):
                    for j2 in range(j, jtar):
                        K_view[i2, j2] += K_ij[i2 - i, j2 - j] * num_coopts

    # store the newly computed K matrix
    Ks[(k, l)] = K

###############################################
# Standard Edit Distance with Kronecker Delta #
###############################################

def standard_ted(x_nodes, x_adj, y_nodes = None, y_adj = None):
    """ Computes the standard tree edit distance between the trees x and y,
    each described by a list of nodes and an adjacency list adj, where adj[i]
    is a list of indices pointing to children of node i.

    The 'standard' refers to the fact that we use the kronecker distance
    as delta, i.e. this call computes the same as

    ted(x_nodes, x_adj, y_nodes, y_adj, kronecker_distance) where

    kronecker_distance(x, y) = 1 if x != y and 0 if x = y.

    However, this implementation here is notably faster because we can apply
    integer arithmetic.

    Note that we assume a proper depth-first-search order of adj, i.e. for
    every node i, the following indices are all part of the subtree rooted at
    i until we hit the index of i's right sibling or the end of the tree.

    Parameters
    ----------
    x_nodes: list or tuple
        a list of nodes for tree x OR a tuple of the form (x_nodes, x_adj).
    x_adj: list or tuple
        an adjacency list for tree x OR a tuple of the form (y_nodes, y_adj).
    y_nodes: list (default = x_adj[0])
        a list of nodes for tree y.
    y_adj: list (default = x_adj[1])
        an adjacency list for tree y.

    Returns
    -------
    d: int
        the standard tree edit distance between x and y according.

    """
    if(isinstance(x_nodes, tuple)):
        x_nodes, x_adj, y_nodes, y_adj = extract_from_tuple_input(x_nodes, x_adj)

    # the number of nodes in both trees
    cdef int m = len(x_nodes)
    cdef int n = len(y_nodes)
    # if the left tree is empty, the standard edit distance is n, and vice
    # versa
    if(m == 0):
        return n
    if(n == 0):
        return m

    _, _, _, _, _, _, D_tree = _standard_ted(x_nodes, x_adj, y_nodes, y_adj)
    return D_tree[0,0]

def _standard_ted(x_nodes, x_adj, y_nodes = None, y_adj = None):
    """ Internal function; call standard_ted instead. """
    if(isinstance(x_nodes, tuple)):
        x_nodes, x_adj, y_nodes, y_adj = extract_from_tuple_input(x_nodes, x_adj)

    # the number of nodes in both trees
    cdef int m = len(x_nodes)
    cdef int n = len(y_nodes)
    # An array to store which pairs of symbols in x and y are equal
    Delta = np.zeros((m, n), dtype=int)
    cdef long[:,:] Delta_view = Delta
    cdef int i
    cdef int j
    for i in range(m):
        for j in range(n):
            if(x_nodes[i] != y_nodes[j]):
                Delta_view[i,j] = 1

    # Compute the keyroots and outermost right leaves for both trees.
    x_orl = outermost_right_leaves(x_adj)
    x_kr  = keyroots(x_orl)
    y_orl = outermost_right_leaves(y_adj)
    y_kr  = keyroots(y_orl)

    # Finally, compute the actual tree edit distance
    D_forest = np.zeros((m+1,n+1), dtype=int)
    D_tree = np.zeros((m,n), dtype=int)
    _std_ted_c(x_orl, x_kr, y_orl, y_kr, Delta, D_forest, D_tree)

    return x_orl, x_kr, y_orl, y_kr, Delta, D_forest, D_tree

@cython.boundscheck(False)
cdef void _std_ted_c(const long[:] x_orl, const long[:] x_kr, const long[:] y_orl, const long[:] y_kr, const long[:,:] Delta, long[:,:] D, long[:,:] D_tree) nogil:
    """ This method is internal and performs the actual standard tree edit
    distance computation for trees x and y in pure C.

    For details on the algorithm, please refer to the following tutorial:
    https://arxiv.org/abs/1805.06869

    Let m and n be the size of tree x and y respectively.

    Parameters
    ----------
    x_orl: long array
        the outermost right leaves for tree x (int array of length m).
    x_kr: long array
        the keyroots for tree x in descending order (int array).
    y_orl: long array
        the outermost right leaves for tree y (int array of length n).
    y_kr: long array
        the keyroots for tree y in descending order (int array).
    Delta: long matrix
        an (m+1) x (n+1) matrix, where Delta[i,j] for i < m, j < n is the
        cost of replacing x[i] with y[j], where Delta[i,n] is the cost of
        deleting x[i], and where Delta[m,j] is the cost of inserting y[j].
    D: long matrix
        an empty (m+1) x (n+1) matrix used for temporary computations.
    D_tree: long matrix
        an empty m x n matrix. After this method has run, D_tree[i,j] will
        be the tree edit distance between the subtree rooted at i and the
        subtree rooted at j.

    """
    # the number of nodes in both trees
    cdef int m = len(x_orl)
    cdef int n = len(y_orl)
    # the number of keyroots in both trees
    cdef int K = len(x_kr)
    cdef int L = len(y_kr)

    # set up iteration variables
    # for the keyroots
    cdef int k
    cdef int l
    # for the nodes in the subtrees rooted at the keyroots
    cdef long i
    cdef long j
    # and temporary variables for the keyroots and the outermost right leaves
    cdef long i_0
    cdef long j_0
    cdef long i_max
    cdef long j_max

    # iterate over all pairwise combinations of keyroots
    for k in range(K):
        for l in range(L):
            # We consider now the subtree rooted at x_kr[k] versus the subtree
            # rooted at y_kr[l]. The forest edit distances between these
            # subtrees correspond exactly to the matrix block
            # D[x_kr[k]:x_orl[x_kr[k]]+1, y_kr[l]:y_orl[y_kr[l]]+1],
            # which we compute now.
            i_0 = x_kr[k]
            j_0 = y_kr[l]
            i_max = x_orl[i_0] + 1
            j_max = y_orl[j_0] + 1
            # first, initialize the last entry for the current subtree
            # computation
            D[i_max, j_max] = 0
            # then, initialize the last column
            for i in range(i_max-1, i_0-1, -1):
                D[i, j_max] = 1 + D[i+1, j_max]
            # then, initialize the last row
            for j in range(j_max-1, j_0-1, -1):
                D[i_max, j] = 1 + D[i_max, j+1]
            # finally, compute the remaining forest edit distances
            for i in range(i_max-1, i_0-1, -1):
                for j in range(j_max-1, j_0-1, -1):
                    if(x_orl[i] == x_orl[i_0] and y_orl[j] == y_orl[j_0]):
                        # if we consider a complete subtree, the forest edit
                        # distance D[i,j] is equal to the tree edit distance
                        # at that position and we can compute it via the
                        # standard edit distance recurrence
                        D[i,j] = min3_int(Delta[i,j] + D[i+1,j+1], # replacement
                                      1 + D[i+1,j], # deletion
                                      1 + D[i,j+1] # insertion
                                 )
                        # store the newly computed tree edit distance as well
                        D_tree[i,j] = D[i,j]
                    else:
                        # if we do _not_ consider a complete subtree, replacements
                        # are only possible between entire subtrees, which we have
                        # to consider in recurrence
                        D[i,j] = min3_int(D_tree[i,j] + D[x_orl[i]+1,y_orl[j]+1], # tree replacement
                                      1 + D[i+1,j], # deletion
                                      1 + D[i,j+1] # insertion
                                 )

cdef long min3_int(long a, long b, long c) nogil:
    """ Computes the minimum of three numbers.

    Parameters
    ----------
    a: int
        a number
    b: int
        another number
    c: int
        yet another number

    Returns
    -------
    min3: int
        min({a, b, c})

    """
    if(a < b):
        if(a < c):
            return a
        else:
            return c
    else:
        if(b < c):
            return b
        else:
            return c

#########################
# Backtracing Functions #
#########################

def standard_ted_backtrace(x_nodes, x_adj, y_nodes = None, y_adj = None):
    """ Computes the standard tree edit distance between the trees x and y,
    each described by a list of nodes and an adjacency list adj, where adj[i]
    is a list of indices pointing to children of node i. This function
    returns an alignment representation of the distance.

    The 'standard' refers to the fact that we use the kronecker distance
    as delta, i.e. this call computes the same as

    ted(x_nodes, x_adj, y_nodes, y_adj, kronecker_distance) where

    kronecker_distance(x, y) = 1 if x != y and 0 if x = y.

    However, this implementation here is notably faster because we can apply
    integer arithmetic.

    Note that we assume a proper depth-first-search order of adj, i.e. for
    every node i, the following indices are all part of the subtree rooted at
    i until we hit the index of i's right sibling or the end of the tree.

    Parameters
    ----------
    x_nodes: list or tuple
        a list of nodes for tree x OR a tuple of the form (x_nodes, x_adj).
    x_adj: list or tuple
        an adjacency list for tree x OR a tuple of the form (y_nodes, y_adj).
    y_nodes: list (default = x_adj[0])
        a list of nodes for tree y.
    y_adj: list (default = x_adj[1])
        an adjacency list for tree y.

    Returns
    -------
    alignment: class alignment.Alignment
        A co-optimal alignment between x and y according to the tree edit
        distance.

    """
    if(isinstance(x_nodes, tuple)):
        x_nodes, x_adj, y_nodes, y_adj = extract_from_tuple_input(x_nodes, x_adj)
    x_orl, x_kr, y_orl, y_kr, Delta, D, D_tree = _standard_ted(x_nodes, x_adj, y_nodes, y_adj)
    # initialize the alignment
    ali = Alignment()
    # start backtracing recursively
    _standard_ted_backtrace(x_orl, y_orl, Delta, D, D_tree, ali, 0, 0)
    return ali

def _standard_ted_backtrace(const long[:] x_orl, const long[:] y_orl, const long[:,:] Delta, long[:,:] D, const long[:,:] D_tree, ali, int k, int l):
    """ Internal function; call standard_ted_backtrace instead.

        Performs the backtracing for the subtree rooted at k in x versus the
        subtree rooted at l in y.
    """
    # recompute the dynamic programming matrix for the current subtree
    # combination
    cdef int i_max = x_orl[k] + 1
    cdef int j_max = y_orl[l] + 1
    cdef int i = 0
    cdef int j = 0
    if(k > 0 or l > 0):
        # if we are either at k > 0 or l > 0, the first action is a replacement
        ali.append_tuple(k, l)
        # note that D[i_max, j_max] is already correctly initialized
        # initialize the last column
        for i in range(i_max-1, k, -1):
            D[i, j_max] = 1 + D[i+1, j_max]
        # then, initialize the last row
        for j in range(j_max-1, l, -1):
            D[i_max, j] = 1 + D[i_max, j+1]
        # finally, compute the remaining forest edit distances
        for i in range(i_max-1, k, -1):
            for j in range(j_max-1, l, -1):
                if(x_orl[i] == x_orl[k] and y_orl[j] == y_orl[l]):
                    # if we consider a complete subtree, we can re-use
                    # the tree edit distance values we computed in the
                    # forward pass
                    D[i,j] = D_tree[i,j] + D[i_max, j_max]
                else:
                    # if we do _not_ consider a complete subtree,
                    # replacements are only possible between entire
                    # subtrees, which we have to consider in
                    # recurrence
                    D[i,j] = min3_int(D_tree[i,j] + D[x_orl[i]+1,y_orl[j]+1], # tree replacement
                                  1 + D[i+1,j], # deletion
                                  1 + D[i,j+1] # insertion
                             )
        i = k + 1
        j = l + 1
    # now, start the backtracing for the current subtree combination
    while(i < i_max and j < j_max):
        # check whether a deletion is co-optimal
        if(D[i, j] == 1 + D[i+1, j]):
            # if so, append a deletion operation, increment i, and continue
            ali.append_tuple(i, -1)
            i += 1
            continue
        # check whether an insertion is co-optimal
        if(D[i, j] == 1 + D[i, j+1]):
            # if so, append an insertion operation, increment j, and continue
            ali.append_tuple(-1, j)
            j += 1
            continue
        # check wehther replacement is co-optimal. In this case, we need to
        # consider two cases
        if(x_orl[i] == x_orl[k] and y_orl[j] == y_orl[l]):
            # If we are at the root of postfix-subtrees for subtree k and l,
            # we consider the standard replacement case
            if(D[i,j] == Delta[i,j] + D[i+1,j+1]):
                # append a replacement operation, increment i and j, and
                # continue
                ali.append_tuple(i, j)
                i += 1
                j += 1
                continue
        else:
            if(D[i,j] == D_tree[i,j] + D[x_orl[i]+1,y_orl[j]+1]):
                # Otherwise, we consider the case where we replace the entire
                # subtree rooted at i with the entire subtree rooted at j.
                # For this case, we call the backtracing recursively
                _standard_ted_backtrace(x_orl, y_orl, Delta, D, D_tree, ali, i, j)
                i = x_orl[i]+1
                j = y_orl[j]+1
                continue
        # if we got here, nothing is co-optimal, which is an error
        raise ValueError('Internal error: No option is co-optimal.')
    # delete and insert any remaining nodes
    while(i < i_max):
        ali.append_tuple(i, -1)
        i += 1
    while(j < j_max):
        ali.append_tuple(-1, j)
        j += 1

def standard_ted_backtrace_matrix(x_nodes, x_adj, y_nodes = None, y_adj = None):
    """ Computes a matrix P where entry P[i, j] represents how often node
    i in tree x was aligned with node j in tree y in co-optimal alignments
    according to the standard tree edit distance.

    Note that we assume a proper depth-first-search order of adj, i.e. for
    every node i, the following indices are all part of the subtree rooted at
    i until we hit the index of i's right sibling or the end of the tree.

    Parameters
    ----------
    x_nodes: list or tuple
        a list of nodes for tree x OR a tuple of the form (x_nodes, x_adj).
    x_adj: list or tuple
        an adjacency list for tree x OR a tuple of the form (y_nodes, y_adj).
    y_nodes: list (default = x_adj[0])
        a list of nodes for tree y.
    y_adj: list (default = x_adj[1])
        an adjacency list for tree y.

    Returns
    -------
    P: array_like
        a matrix, where entry P[i, j] specifies the fraction of co-optimal
        alignments in which node x[i] has been aligned with node y[j].
        P[i, n] contains the fraction of deletions of node x[i] and P[m, j]
        the fraction of insertions of node y[j].
    K: array_like
        a matrix that contains the counts for all co-optimal alignments in
        which node x[i] has been aligned with node y[j].
    k: int
        the number of co-optimal alignments overall, such that P = K / k.

    """
    if(isinstance(x_nodes, tuple)):
        x_nodes, x_adj, y_nodes, y_adj = extract_from_tuple_input(x_nodes, x_adj)

    m = len(x_nodes)
    n = len(y_nodes)
    # compute tree edit distance first
    x_orl, x_kr, y_orl, y_kr, Delta, D, D_tree = _standard_ted(x_nodes, x_adj, y_nodes, y_adj)

    # set up a dictionary to sparsely store the counting matrices for all subtrees
    Ks = {}
    # set up a matrix to store the number of co-optimal alignments for all subtrees
    Kappa = np.zeros((m, n), dtype=int)

    # start the recursive backtrace computation
    _standard_ted_backtrace_matrix(x_orl, x_kr, y_orl, y_kr, Delta, D, D_tree, Ks, Kappa, 0, 0)

    # extract results
    K = Ks[(0,0)]
    k = Kappa[0, 0]
    # construct P
    P = np.zeros((m+1, n+1))
    P[:m, :][:, :n] = K
    P[:m, n] = k - np.sum(K, axis=1)
    P[m, :n] = k - np.sum(K, axis=0)
    P /= k
    # return results
    return P, K, k

def _standard_ted_backtrace_matrix(const long[:] x_orl, const long[:] x_kr, const long[:] y_orl, const long[:] y_kr, const long[:,:] Delta, long[:,:] D, const long[:,:] D_tree, Ks, long[:,:] Kappa, int k, int l):
    """ Internal function; call standard_ted_backtrace_matrix instead.

        Performs the backtracing for the subtree rooted at k in x versus the
        subtree rooted at l in y.
    """

    # get the sizes of the current subtrees
    cdef int m_k = x_orl[k] - k + 1
    cdef int n_l = y_orl[l] - l + 1
    if(m_k == 1 and n_l == 1):
        # if this is a pair of leaves, the handling is trivial
        Kappa[k, l] = 1
        Ks[(k,l)] = np.array([[1]], dtype=int)
        return

    cdef int m = len(x_orl)
    cdef int n = len(y_orl)

    # compute the forward matrix Alpha, which contains the number of
    # co-optimal alignment paths from cell [k, l, k, l] to cell [k, l, i, j]
    Alpha = np.zeros((m_k+1, n_l+1), dtype=int)
    cdef long[:, :] Alpha_view = Alpha
    # we start with a single path
    Alpha_view[0, 0] = 1
    cdef int i
    cdef int j
    cdef long[:,:] D_kl
    if(k == 0 and l == 0):
        D_kl = D
    else:
        D_kl = np.zeros((m_k+1, n_l+1), dtype=int)

    if(k > 0 or l > 0):
        # if we are not considering the entire tree, we always use a
        # replacement as first action to avoid duplicates with other options
        Alpha_view[1, 1] = 1
        # re-compute the edit costs for the current subtree combination
        D_kl[m_k][n_l] = D[x_orl[k], y_orl[l]]
        D_kl[0][0]     = D_tree[k][l] + D[x_orl[k], y_orl[l]]
        for i in range(m_k-1, 0, -1):
            D_kl[i, n_l] = 1 + D_kl[i+1, n_l]
        # then, initialize the last row
        for j in range(n_l-1, 0, -1):
            D_kl[m_k, j] = 1 + D_kl[m_k, j+1]
        # finally, compute the remaining forest edit distances
        for i in range(m_k-1, 0, -1):
            for j in range(n_l-1, 0, -1):
                if(x_orl[k+i] == x_orl[k] and y_orl[l+j] == y_orl[l]):
                    # if we consider a complete subtree, we can re-use
                    # the tree edit distance values we computed in the
                    # forward pass
                    D_kl[i,j] = D_tree[k+i,l+j] + D_kl[m_k][n_l]
                else:
                    # if we do _not_ consider a complete subtree,
                    # replacements are only possible between entire
                    # subtrees, which we have to consider in
                    # recurrence
                    D_kl[i,j] = min3_int(D_tree[k+i,l+j] + D_kl[x_orl[k+i]-k+1,y_orl[l+j]-l+1], # tree replacement
                                  1 + D_kl[i+1,j], # deletion
                                  1 + D_kl[i,j+1]  # insertion
                             )
        # start our processing queue at 1,1
        q = [(1, 1)]
    else:
        # start our processing queue at 0,0
        q = [(0, 0)]
    # build a set which stores the already visited cells
    visited = set()
    # search along co-optimal paths until the processing queue is empty
    cdef int found_coopt
    cdef int itar
    cdef int jtar
    while(q):
        (i, j) = heapq.heappop(q)
        if((i, j) in visited):
            continue
        visited.add((i, j))
        num_coopts = Alpha_view[i, j]
        if(i == m_k):
            if(j == n_l):
                continue
            # if we are at the end of the first subtree, we can only insert
            if(D_kl[i, j] == 1 + D_kl[i,j+1]):
                Alpha_view[i, j+1] += num_coopts
                heapq.heappush(q, (i, j+1))
            else:
                raise ValueError('Internal error: No option is co-optimal.')
            continue
        if(j == n_l):
            # if we are at the end of the second subtree, we can only delete
            if(D_kl[i, j] == 1 + D_kl[i+1, j]):
                Alpha_view[i+1, j] += num_coopts
                heapq.heappush(q, (i+1, j))
            else:
                raise ValueError('Internal error: No option is co-optimal.')
            continue

        found_coopt = False
        if(D_kl[i, j] == 1 + D_kl[i+1, j]):
            # deletion is co-optimal
            Alpha_view[i+1, j] += num_coopts
            heapq.heappush(q, (i+1, j))
            found_coopt = True
        if(D_kl[i, j] == 1 + D_kl[i,j+1]):
            # insertion is co-optimal
            Alpha_view[i, j+1] += num_coopts
            heapq.heappush(q, (i, j+1))
            found_coopt = True
        if(x_orl[k+i] == x_orl[k] and y_orl[l+j] == y_orl[l]):
            # If we are at the root of postfix-subtrees for subtree k and l,
            # we consider the standard replacement case
            if(D_kl[i,j] == Delta[k+i,l+j] + D_kl[i+1,j+1]):
                # replacement is co-optimal
                Alpha_view[i+1, j+1] += num_coopts
                heapq.heappush(q, (i+1, j+1))
                found_coopt = True
        else:
            itar = x_orl[k+i]-k+1
            jtar = y_orl[l+j]-l+1
            if(D_kl[i, j] == D_tree[k+i,l+j] + D_kl[itar, jtar]):
                # Otherwise, we consider the case where we replace the entire
                # subtree rooted at i with the entire subtree rooted at j.
                # For this case, we call the backtracing recursively
                if((k+i, l+j) not in Ks):
                    _standard_ted_backtrace_matrix(x_orl, x_kr, y_orl, y_kr, Delta, D, D_tree, Ks, Kappa, k+i, l+j)
                # then, we can use the number of paths during recursion,
                # multiplied with the number of coopts we have accumulated so
                # far
                Alpha_view[itar][jtar] += num_coopts * Kappa[k+i,l+j]
                heapq.heappush(q, (itar, jtar))
                found_coopt = True
        if(not found_coopt):
            raise ValueError('Internal error: No option is co-optimal.')

    # store the number of co-optimals for this subtree
    Kappa[k, l] = Alpha_view[m_k, n_l]

    # next, we compute the backward counting matrix, Beta, which contains the
    # number of co-optimal alignment paths from cell [k, l, i, j] to cell
    # [k, l, m_k, n_l]
    Beta = np.zeros((m_k+1, n_l+1), dtype=int)
    cdef long[:, :] Beta_view = Beta
    Beta_view[m_k, n_l] = 1
    # add (0,0) to the visited set to ensure consistency because we started
    # at (1, 1) during forward computation
    visited.add((0,0))
    # iterate in downward lexigraphic order over the visited cells
    for (i, j) in sorted(visited, reverse = True):
        if(i == m_k):
            if(j == n_l):
                continue
            # if we are at the end of the first subtree, we can only insert
            if(D_kl[i, j] == 1 + D_kl[i,j+1]):
                Beta_view[i, j] += Beta_view[i, j+1]
            else:
                raise ValueError('Internal error: No option is co-optimal.')
            continue
        if(j == n_l):
            # if we are at the end of the second subtree, we can only delete
            if(D_kl[i, j] == 1 + D_kl[i+1, j]):
                Beta_view[i, j] += Beta_view[i+1,j]
            else:
                raise ValueError('Internal error: No option is co-optimal.')
            continue

        found_coopt = False
        if(D_kl[i, j] == 1 + D_kl[i+1, j]):
            # deletion is co-optimal
            Beta_view[i, j] += Beta_view[i+1, j]
            found_coopt = True
        if(D_kl[i, j] == 1 + D_kl[i,j+1]):
            # insertion is co-optimal
            Beta_view[i, j] += Beta_view[i, j+1]
            found_coopt = True
        if(x_orl[k+i] == x_orl[k] and y_orl[l+j] == y_orl[l]):
            # If we are at the root of postfix-subtrees for subtree k and l,
            # we consider the standard replacement case
            if(D_kl[i,j] == Delta[k+i,l+j] + D_kl[i+1,j+1]):
                # replacement is co-optimal
                Beta_view[i, j] += Beta_view[i+1, j+1]
                found_coopt = True
        else:
            itar = x_orl[k+i]-k+1
            jtar = y_orl[l+j]-l+1
            if(D_kl[i, j] == D_tree[k+i,l+j] + D_kl[itar, jtar]):
                # if we replace an entire subtree, we need to consider the
                # number of co-optimal alignments between those, which is
                # listed in Kappa
                Beta_view[i][j] += Beta_view[itar][jtar] * Kappa[k+i,l+j]
                found_coopt = True
        if(not found_coopt):
            raise ValueError('Internal error: No option is co-optimal.')

    if(Alpha_view[m_k, n_l] != Beta_view[0, 0]):
        raise ValueError('Internal error: Alignment count in Alpha and Beta matrix did not agree; got %d versus %d' % (Alpha_view[m_k, n_l], Beta_view[0, 0]))

    # initialize the counting matrix for the current subtree
    K = np.zeros((m_k, n_l), dtype=int)
    cdef long[:,:] K_view = K
    cdef long[:,:] K_ij
    cdef int i2
    cdef int j2
    # compute content of K
    for (i, j) in visited:
        if(i == m_k or j == n_l):
            continue
        if(x_orl[k+i] == x_orl[k] and y_orl[l+j] == y_orl[l]):
            # If we are at the root of postfix-subtrees for subtree k and l
            # _or_ if replacements are as expensive as deletions plus insertions,
            # we count replacements directly
            if(D_kl[i,j] == Delta[k+i,l+j] + D_kl[i+1,j+1]):
                K_view[i, j] += Alpha_view[i,j] * Beta_view[i+1,j+1]
        else:
            itar = x_orl[k+i]-k+1
            jtar = y_orl[l+j]-l+1
            if(D_kl[i, j] == D_tree[k+i,l+j] + D_kl[itar, jtar]):
                # if we replace an entire subtree, we need to consider the
                # number of co-optimal alignments between those, which is
                # listed in Kappa
                num_coopts = Alpha_view[i,j] * Beta_view[itar,jtar]
                K_ij = Ks[(k+i, l+j)]
                for i2 in range(i, itar):
                    for j2 in range(j, jtar):
                        K_view[i2, j2] += K_ij[i2 - i, j2 - j] * num_coopts

    # store the newly computed K matrix
    Ks[(k, l)] = K



"""
Implements list edits, i.e. functions which take a list as input and
return a changed list.

"""
# Copyright (C) 2019-2021
# Benjamin Paaßen
# AG Machine Learning
# Bielefeld University

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
import copy

__author__ = 'Benjamin Paaßen'
__copyright__ = 'Copyright (C) 2019-2021, Benjamin Paaßen'
__license__ = 'GPLv3'
__version__ = '1.2.0'
__maintainer__ = 'Benjamin Paaßen'
__email__  = 'bpaassen@techfak.uni-bielefeld.de'

class Edit(abc.ABC):
    """ An abstract parent class for all edits.

    """
    @abc.abstractmethod
    def apply(self, lst):
        """ Applies this edit to the given list and returns a copy of the
        list with the applied changes. The original list remains unchanged.

        Parameters
        ----------
        lst: list
            a list

        Returns
        -------
        res: list
            a copy of the list with the applied edit.

        """
        pass

    @abc.abstractmethod
    def apply_in_place(self, lst):
        """ Applies this edit to the given list. Note that this changes the
        input argument.

        Parameters
        ----------
        lst: list
            a list

        """
        pass


class Replacement(Edit):
    """ Replaces node self._index with label self._label.

    Attributes
    ----------
    _index: int
        The index of the node to be deleted.
    _label: str
        The new label for node self._index.

    """
    def __init__(self, index, label):
        self._index = index
        self._label = label

    def apply(self, lst):
        """ Replaces entry self._index with self._label.

        Parameters
        ----------
        lst: list
            a list

        Returns
        -------
        res: list
            a copy of the list with the applied edit.

        """
        lst = copy.copy(lst)
        self.apply_in_place(lst)
        return lst

    def apply_in_place(self, lst):
        """ Replaces entry self._index with self._label.

        Parameters
        ----------
        lst: list
            a list

        """
        lst[self._index] = self._label

    def __repr__(self):
        return 'rep(%d, %s)' % (self._index, str(self._label))

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        return isinstance(other, Replacement) and self._index == other._index and self._label == other._label


class Deletion(Edit):
    """ Deletes node self._index.

    Attributes
    ----------
    _index: int
        The index of the node to be deleted.

    """
    def __init__(self, index):
        self._index = index

    def apply(self, lst):
        """ Deletes the self._indexth entry.

        Parameters
        ----------
        lst: list
            a list

        Returns
        -------
        res: list
            a copy of the list with the applied edit.

        """
        lst = copy.copy(lst)
        self.apply_in_place(lst)
        return lst

    def apply_in_place(self, lst):
        """ Deletes the self._indexth entry.

        Parameters
        ----------
        lst: list
            a list

        """
        # delete the nodes entry at index
        del lst[self._index]

    def __repr__(self):
        return 'del(%d)' % (self._index)

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        return isinstance(other, Deletion) and self._index == other._index


class Insertion(Edit):
    """ Inserts a new self._label entry at position self._index .

    Attributes
    ----------
    _index: int
        The index at which this edit will be applied.
    _label: str
        The new entry.

    """
    def __init__(self, index, label):
        self._index = index
        self._label = label

    def apply(self, lst):
        """ Inserts a new self._label entry at position self._index.

        Parameters
        ----------
        lst: list
            a list

        Returns
        -------
        res: list
            a copy of the list with the applied edit.

        """
        lst = copy.copy(lst)
        self.apply_in_place(lst)
        return lst

    def apply_in_place(self, lst):
        """ Inserts a new self._label entry at position self._index.

        Parameters
        ----------
        lst: list
            a list

        """
        lst.insert(self._index, self._label)

    def __repr__(self):
        return 'ins(%d, %s)' % (self._index, str(self._label))

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        return isinstance(other, Insertion) and self._index == other._index and self._label == other._label


class Script(list, Edit):
    """ A list of Edits.

    """
    def __init__(self, lst = []):
        list.__init__(self, lst)

    def apply(self, lst):
        """ Applies all edits in this script.

        Parameters
        ----------
        lst: list
            a list

        Returns
        -------
        res: list
            a copy of the list with the applied edits of this script.

        """
        # if this script is empty, just return the input
        if(not self):
            return lst
        # otherwise make a copy and then apply the script in place
        lst = copy.copy(lst)
        self.apply_in_place(lst)
        return lst

    def apply_in_place(self, lst):
        """ Applies all edits in this script in place.

        Parameters
        ----------
        lst: list
            a list

        """
        for e in range(len(self)):
            self[e].apply_in_place(lst)


def alignment_to_script(alignment, x, y):
    """ Converts the given alignment into an edit script which maps the given
    list x to the given list y such that all entries of the alignment are
    translated one to one into edits.

    Note that the order of operations does change because the script will
    first apply replacements (in input order), then deletions
    (in descending order), and finally insertions (in ascending order),
    which simplifies conversion.

    Parameters
    ----------
    alignment: class alignment.Alignment
        an Alignment object which maps between the given lists x and y.
    x: list
        a list.
    y: list
        another list.

    Returns
    -------
    script: class edits.Script
        A Script object script such that script.apply(x) = y and where every
        tuple in the alignment has one corresponding edit.

    """
    script = Script()
    # iterate over the alignment and sort replacements, deletions, and
    # insertions
    dels = []
    inss = []
    for op in alignment:
        if(op._left >= 0):
            if(op._right >= 0):
                # if both left and right are >= 0, we have a replacement,
                # which we can process right away
                # We ignore replacements that change nothing, though.
                if(x[op._left] != y[op._right]):
                    script.append(Replacement(op._left, y[op._right]))
            else:
                # if only the left index is defined, we have a deletion, which
                # we process later
                dels.append(op._left)
        else:
            if(op._right < 0):
                raise ValueError('Invalid tuple in alignment: %s.' % str(op))
            # if only the right index is defined, we have an insertion, which
            # we process later
            inss.append(op._right)
    # apply deletions in reverse order
    for i in sorted(dels, reverse=True):
        script.append(Deletion(i))
    # apply insertions in ascending order
    for j in sorted(inss):
        script.append(Insertion(j, y[j]))
    # return the script
    return script



#!python
#cython: language_level=3
"""
Implements the constrained unordered tree edit distance of Zhang (1996) and its
backtracing in cython.

"""
# Copyright (C) 2021
# Benjamin Paaßen
# AG Machine Learning
# Bielefeld University

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from edist.ted import extract_from_tuple_input
import numpy as np
from scipy.optimize import linear_sum_assignment
from edist.alignment import Alignment
cimport cython

__author__ = 'Benjamin Paaßen'
__copyright__ = 'Copyright 2021, Benjamin Paaßen'
__license__ = 'GPLv3'
__version__ = '0.1.0'
__maintainer__ = 'Benjamin Paaßen'
__email__  = 'bpaassen@techfak.uni-bielefeld.de'


cdef extern from "math.h":
    float INFINITY

###################################
# Edit Distance with Custom Delta #
###################################

def uted(x_nodes, x_adj, y_nodes = None, y_adj = None, delta = None):
    """ Computes the constrained, unordered tree edit distance between the
     trees x and y, each described by a list of nodes and an adjacency list
    adj, where adj[i] is a list of indices pointing to children of node i.

    Note that we assume a proper depth-first-search order of adj, i.e. for
    every node i, the following indices are all part of the subtree rooted at
    i until we hit the index of i's right sibling or the end of the tree.

    Parameters
    ----------
    x_nodes: list or tuple
        a list of nodes for tree x OR a tuple of the form (x_nodes, x_adj).
    x_adj: list or tuple
        an adjacency list for tree x OR a tuple of the form (y_nodes, y_adj).
    y_nodes: list (default = x_adj[0])
        a list of nodes for tree y.
    y_adj: list (default = x_adj[1])
        an adjacency list for tree y.
    delta: function (default = None)
        a function that takes two nodes as inputs and returns their pairwise
        distance, where delta(x, None) should be the cost of deleting x and
        delta(None, y) should be the cost of inserting y. If undefined, this
        method uses unit costs.

    Returns
    -------
    d: float
        the tree edit distance between x and y according to delta.

    """
    if isinstance(x_nodes, tuple):
        x_nodes, x_adj, y_nodes, y_adj = extract_from_tuple_input(x_nodes, x_adj)

    # the number of nodes in both trees
    cdef int m = len(x_nodes)
    cdef int n = len(y_nodes)

    # if either tree is empty, we can only delete/insert all nodes in the
    # non-empty tree.
    cdef double d = 0
    cdef int i
    cdef int j
    if m == 0:
        if delta is None:
            return n
        else:
            for j in range(n):
                d += delta(None, y_nodes[j])
            return d
    if n == 0:
        if delta is None:
            return m
        else:
            for i in range(m):
                d += delta(x_nodes[i], None)
            return d

    # Set up an array to store edit costs for replacements,
    # deletions, and insertions
    Delta = np.ones((m+1, n+1))
    cdef double[:,:] Delta_view = Delta

    if delta is None:
        for i in range(m):
            for j in range(n):
                if x_nodes[i] == y_nodes[j]:
                    Delta_view[i, j] = 0.
    else:
        # First, compute all pairwise replacement costs
        for i in range(m):
            for j in range(n):
                Delta_view[i,j] = delta(x_nodes[i], y_nodes[j])
        # Then, compute the deletion and insertion costs
        for i in range(m):
            Delta_view[i,n] = delta(x_nodes[i], None)
        for j in range(n):
            Delta_view[m,j] = delta(None, y_nodes[j])

    # compute the actual tree edit distance
    D_forest, D_tree = _uted(x_nodes, x_adj, y_nodes, y_adj, Delta)

    return D_tree[0,0]

def _uted(x_nodes, x_adj, y_nodes, y_adj, Delta):
    """ Internal function; call uted instead. """
    # the number of nodes in both trees
    cdef int m = len(x_nodes)
    cdef int n = len(y_nodes)

    # convert adjacency lists to array form
    A_x, deg_x = adjmat_(x_adj)
    A_y, deg_y = adjmat_(y_adj)

    # set up temporary matrices for the munkres algorithm
    cdef int max_deg = A_x.shape[1] + A_y.shape[1]
    C = np.zeros((max_deg, max_deg))
    Stars  = np.zeros((max_deg, max_deg), dtype=np.intc)
    Primes = np.zeros((max_deg, max_deg), dtype=np.intc)
    row_covers = np.zeros(max_deg, dtype=np.intc)
    col_covers = np.zeros(max_deg, dtype=np.intc)
    path = np.zeros((max_deg, 2), dtype=int)
    pi = np.zeros(max_deg, dtype=int)

    # initialize dynamic programming matrices for forest and tree edit distance
    D_forest = np.zeros((m+1,n+1))
    D_tree = np.zeros((m+1,n+1))
    
    # call the c routine
    uted_c_(A_x, deg_x, A_y, deg_y, Delta, D_forest, D_tree, C, Stars, Primes, row_covers, col_covers, path, pi)

    return D_forest, D_tree

def adjmat_(adj):
    """ Converts an adjacency list into an int array """
    cdef int m = len(adj)
    degs = np.zeros(m, dtype=int)
    cdef long[:] deg_view = degs
    cdef int i
    for i in range(m):
        deg_view[i] = len(adj[i])

    cdef long max_deg = np.max(degs)
    Adj = np.zeros((m, max_deg), dtype=int)
    cdef long[:, :] Adj_view = Adj

    cdef int k
    for i in range(m):
        for k in range(deg_view[i]):
            Adj_view[i, k] = adj[i][k]

    return Adj, degs


@cython.boundscheck(False)
cdef void uted_c_(const long[:,:] A_x, const long[:] deg_x, const long[:,:] A_y, const long[:] deg_y, 
    const double[:,:] Delta, double[:,:] D_forest, double[:,:] D_tree,
    double[:, :] C, int[:, :] Stars, int[:, :] Primes,
    int[:] row_covers, int[:] col_covers, long[:, :] path, long[:] pi) nogil:
    """ This method is internal and performs the actual tree edit distance
    computation for trees x and y in pure C.

    For details on the algorithm, please refer to the paper
    'A Constrained Edit Distance Between Unordered Labeled Trees'
    by Zhang (1996).

    """
    # the number of nodes in both trees
    cdef int m = A_x.shape[0]
    cdef int n = A_y.shape[0]

    # for the nodes in x and y
    cdef int i
    cdef int j
    # for the children of x[i] and y[j]
    cdef int k
    cdef int l
    # for the number of children of i and j
    cdef int m_i
    cdef int n_j
    # and for the child indices
    cdef int i_k = 0
    cdef int j_l = 0
    # and for the deletion, insertion, and replacement cost
    cdef double del_cost
    cdef double ins_cost
    cdef double rep_cost

    cdef double tmp_cost

    for i in range(m-1, -1, -1):
        # Compute the subforest deletion cost for i, i.e. the cost
        # for deleting all of i's child subtrees
        for k in range(deg_x[i]):
            i_k = A_x[i, k]
            D_forest[i, n] += D_tree[i_k, n]
        # Deleting the tree rooted at i means deleting node i and all its
        # children
        D_tree[i, n] = Delta[i, n] + D_forest[i, n]

    for j in range(n-1, -1, -1):
        # Compute the subforest insertion cost for j, i.e. the cost
        # for inserting all of j's child subtrees
        for l in range(deg_y[j]):
            j_l = A_y[j, l]
            D_forest[m, j] += D_tree[m, j_l]
        # Inserting the tree rooted at j means inserting node j and all its
        # children
        D_tree[m, j] = Delta[m, j] + D_forest[m, j]

    # now, start the actual recursion
    for i in range(m-1, -1, -1):
        for j in range(n-1, -1, -1):
            m_i = deg_x[i]
            n_j = deg_y[j]
            # First, we compute the forest edit distance, i.e. the cost for
            # editing all children of i into all children of j.

            # We consider first the special case that either i or
            # j have no children. Then, the computation is really
            # simple because we can only delete/insert
            if m_i == 0:
                D_forest[i, j] = D_forest[m, j]
            elif n_j == 0:
                D_forest[i, j] = D_forest[i, n]
            else:
                # if both nodes have children, perform the actual computation.
                # For that, we have three options.
                # First, we could delete all children of i except for a single
                # subtree
                del_cost = INFINITY
                for k in range(m_i):
                    i_k = A_x[i, k]
                    # accordingly, we need to consider the cost of editing
                    # the children of node i_k with the children of j,
                    # plus the cost of deleting all other children of i.
                    tmp_cost = D_forest[i_k, j] + D_forest[i, n] - D_forest[i_k, n]
                    if tmp_cost < del_cost:
                        del_cost = tmp_cost
                # Second, we could insert all children of j except for a single
                # subtree
                ins_cost = INFINITY
                for l in range(n_j):
                    j_l = A_y[j, l]
                    # accordingly, we need to consider the cost of editing
                    # the children of node i to the children of j_l,
                    # plus the cost of inserting all other children of j.
                    tmp_cost = D_forest[i, j_l] + D_forest[m, j] - D_forest[m, j_l]
                    if tmp_cost < ins_cost:
                        ins_cost = tmp_cost
                # Third, we could replace, meaning that we optimally match all
                # children of i to all children of j.

                # if there is only one child in one of the forests, we can
                # solve this with a simplified algorithm
                if m_i == 1:
                    # we obtain the total cost by computing the
                    # cost of inserting all children of j,
                    # minus the cost of inserting one child,
                    # but replacing it with i_k
                    rep_cost = INFINITY
                    i_k = A_x[i, 0]
                    for l in range(n_j):
                        j_l = A_y[j, l]
                        tmp_cost = D_tree[i_k, j_l] + D_forest[m, j] - D_tree[m, j_l]
                        if tmp_cost < rep_cost:
                            rep_cost = tmp_cost
                elif n_j == 1:
                    # we obtain the total cost by computing the
                    # cost of deleting all children of i,
                    # minus the cost of deleting one child,
                    # but replacing it with j_k
                    rep_cost = INFINITY
                    j_l = A_y[j, 0]
                    for k in range(m_i):
                        i_k = A_x[i, k]
                        tmp_cost = D_tree[i_k, j_l] + D_forest[i, n] - D_tree[i_k, n]
                        if tmp_cost < rep_cost:
                            rep_cost = tmp_cost
                else:
                    # if there is more than one child on both sides, we use
                    # the Munkres/Hungarian algorithm.

                    # prepare a cost matrix for the Hungarian algorithm
                    for k in range(m_i):
                        i_k = A_x[i, k]
                        for l in range(n_j):
                            j_l = A_y[j, l]
                            # matching ci with cj means editing the ci'th
                            # child of i to the cj'th child of j.
                            C[k, l] = D_tree[i_k, j_l]
                    C[:m_i, n_j:m_i+n_j] = INFINITY
                    for k in range(m_i):
                        # matching c with n_j + c means deleting the
                        # c'th child of i
                        i_k = A_x[i, k]
                        C[k, n_j + k] = D_tree[i_k, n]
                    C[m_i:m_i+n_j, :n_j] = INFINITY
                    for l in range(n_j):
                        # matching m_i + c with c means inserting the
                        # c'th child of j
                        j_l = A_y[j, l]
                        C[m_i + l, l] = D_tree[m, j_l]
                    C[m_i:m_i+n_j, n_j:m_i+n_j] = 0.
                    # solve the linear sum assignment problem for C. The resulting
                    # minimum cost is our replacement cost
                    # print('C before munkres for (%d, %d):\n%s' % (i, j, str(np.asarray(C))))
                    munkres_(C[:m_i+n_j,:m_i+n_j], Stars[:m_i+n_j,:m_i+n_j], Primes[:m_i+n_j,:m_i+n_j],
                    row_covers[:m_i+n_j], col_covers[:m_i+n_j], path[:m_i+n_j, :], pi[:m_i+n_j])
                    # print('pi after munkres: %s' % str(np.asarray(pi)))
                    # reconstruct the cost of the assignment from pi
                    rep_cost = 0.
                    for k in range(m_i):
                        i_k = A_x[i, k]
                        if pi[k] >= n_j:
                            # if pi[k] >= n_j, tree i_k should be deleted
                            rep_cost += D_tree[i_k, n]
                            # print('delete %d for %g' % (i_k, D_tree[i_k, n]))
                        else:
                            # otherwise we replace i_k with j_pi[k]
                            j_l = A_y[j, pi[k]]
                            rep_cost += D_tree[i_k, j_l]
                            # print('replace %d with %d for %g' % (i_k, j_l, D_tree[i_k, j_l]))
                    for l in range(n_j):
                        j_l = A_y[j, l]
                        if pi[m_i + l] < n_j:
                            # if pi[m_i + l] < n_j, tree j_l should be inserted
                            rep_cost += D_tree[m, j_l]
                            # print('insert %d for %g' % (j_l, D_tree[m, j_l]))
                    # print('cost of pi: %g' % rep_cost)
                # compute minimum across deletion, insertion, and replacement
                D_forest[i, j] = min3(del_cost, ins_cost, rep_cost)

            # next, compute the unordered tree edit distance between the
            # subtrees rooted at i and j.
            # Again, we have three options.
            # First, we could delete node i and all subtrees except a
            # single one.
            if m_i == 0:
                # if i has no children, we delete i and insert
                # everything in j
                del_cost = Delta[i, n] + D_tree[m, j]
            else:
                del_cost = INFINITY
                for k in range(m_i):
                    i_k = A_x[i, k]
                    # accordingly, we need to consider the cost of editing
                    # tree i_k into tree j plus the cost of deleting all other
                    # children of i.
                    tmp_cost = D_tree[i_k, j] + D_tree[i, n] - D_tree[i_k, n]
                    if tmp_cost < del_cost:
                        del_cost = tmp_cost
            # Second, we could insert node j and all children of j except
            # for a single one.
            if n_j == 0:
                # if j has no children, we insert j and delete
                # everything in i
                ins_cost = Delta[m, j] + D_tree[i, n]
            else:
                ins_cost = INFINITY
                for l in range(n_j):
                    j_l = A_y[j, l]
                    # accordingly, we need to consider the cost of editing
                    # tree i into tree j_l plus the cost o inserting all other
                    # children of j.
                    tmp_cost = D_tree[i, j_l] + D_tree[m, j] - D_tree[m, j_l]
                    if tmp_cost < ins_cost:
                        ins_cost = tmp_cost
            # Third, we could replace, meaning that we edit node i into
            # node j and all children of i into all children of j
            rep_cost = Delta[i, j] + D_forest[i, j]
            # compute minimum across deletion, insertion, and replacement
            D_tree[i, j] = min3(del_cost, ins_cost, rep_cost)

cdef double min3(double a, double b, double c) nogil:
    """ Computes the minimum of three numbers.

    Parameters
    ----------
    a: double
        a number
    b: double
        another number
    c: double
        yet another number

    Returns
    -------
    min3: double
        min({a, b, c})

    """
    if(a < b):
        if(a < c):
            return a
        else:
            return c
    else:
        if(b < c):
            return b
        else:
            return c

###############################
# Munkres/Hungarian algorithm #
###############################

@cython.boundscheck(False)
cdef void munkres_(double[:, :] C, int[:, :] Stars, int[:, :] Primes,
    int[:] row_covers, int[:] col_covers, long[:, :] path, long[:] pi) nogil:
    """ Implements the munkres algorithm to find a minimal matching
    for the cost matrix C. All other inputs are auxiliary matrices
    and vectors for computation.

    This implementation strictly follows the description of the
    algorithm given by Bourgeois and Lasalle (1971) in their paper
    "An Extension of the Munkres Algorithm for the Assignment Problem
    to Rectangular Matrices".

    The final assignment is stored in pi afterwards.
    """
    cdef int m = C.shape[0]
    cdef int i = 0
    cdef int j = 0
    cdef int k = 0
    cdef int l = 0
    cdef double h
    cdef int t
    cdef int T
    cdef int found_zero

    # initialization: we subtract the row and column minimum
    for i in range(m):
        h = INFINITY
        for j in range(m):
            if C[i, j] < h:
                h = C[i, j]
        for j in range(m):
            C[i, j] -= h
    for j in range(m):
        h = INFINITY
        for i in range(m):
            if C[i, j] < h:
                h = C[i, j]
        for i in range(m):
            C[i, j] -= h
    # print('C after initialization:\n%s' % str(np.asarray(C)))
    # initialize all auxiliary matrices as false
    Stars[:, :]   = 0
    Primes[:, :]  = 0
    row_covers[:] = 0
    col_covers[:] = 0
    # Step 1: inspect all zeros in the matrix. If there is
    # no starred zero in its row or column, star it.
    for i in range(m):
        for j in range(m):
            if C[i, j] < 1E-5 and row_covers[i] <= 0 and col_covers[j] <= 0:
                Stars[i, j]   = 1
                row_covers[i] = 1
                col_covers[j]  = 1
    # print('Stars after initialization\n%s' % str(np.asarray(Stars)))
    while True:
        # uncover all rows but keep columns covered
        row_covers[:] = False
        # Step 2: uncover all rows, but keep columns
        # with a starred zero covered. If that covers
        # all columns, we are done.
        found_zero = 0
        for j in range(m):
            if col_covers[j] <= 0:
                found_zero = 1
                break
        # print('checking step 2: %d' % found_zero)
        if found_zero <= 0:
            # recover the assignment
            for i in range(m):
                for j in range(m):
                    if Stars[i, j] > 0:
                        pi[i] = j
                        break
            # end computation
            return
        # Step 3: Prime non-covered zeros and change coverings
        while True:
            # find an uncovered zero
            found_zero = 0
            # print('row covers in step 3: %s' % str(np.asarray(row_covers)))
            # print('col covers in step 3: %s' % str(np.asarray(col_covers)))
            for i in range(m):
                if row_covers[i] > 0:
                    continue
                for j in range(m):
                    if col_covers[j] > 0:
                        continue
                    if C[i, j] < 1E-5:
                        found_zero = 1
                        break
                if found_zero > 0:
                    break
            # print('found zero in step 3: %d' % found_zero)
            if found_zero > 0:
                Primes[i, j] = 1
                # print('primed (%d, %d) in step 3' % (i, j))
                # search for a zero in the same row
                for l in range(m):
                    if Stars[i, l] > 0:
                        break
                if Stars[i, l]:
                    # print('covered row %d and uncovered column %d in step 3' % (i, l))
                    # if there is a starred zero in the same row
                    # as a primed zero, cover the row and uncover
                    # the column of the starred zero
                    row_covers[i] = 1
                    col_covers[l] = 0
                else:
                    # break and go to step 4
                    break
            else:
                # step 5
                # find non-covered minimum
                h = INFINITY
                for j in range(m):
                    if col_covers[j] > 0:
                        continue
                    for i in range(m):
                        if row_covers[i] > 0:
                            continue
                        if C[i, j] < h:
                            h = C[i, j]
                # add it to covered rows
                for i in range(m):
                    if row_covers[i] > 0:
                        for j in range(m):
                            C[i, j] += h
                # and uncovered columns
                for j in range(m):
                    if col_covers[j] > 0:
                        continue
                    for i in range(m):
                        C[i, j] -= h
                # print('new C after step 5\n%s' % str(np.asarray(C)))
                # conclude step 5, return to step 3

        # step 3 loop is completed, start step 4

        # step 4. We should now find a sequence of
        # starred and primed zeros by alternating
        # rows and columns.
        k = i
        l = j
        # print('Found prime at (%d, %d) in step 4' % (k, l))
        path[0, 0] = k
        path[0, 1] = l
        T = 1
        # continue the path as far as possible
        while True:
            # find the next star in the same column
            for k in range(m):
                if Stars[k, l] > 0:
                    break
            if Stars[k, l] <= 0:
                # if there is no star, the path is complete
                break
            # print('Found star at (%d, %d) in step 4' % (k, l))
            # if there is a star, remove it
            Stars[k, l] = 0
            # then, find the next prime in the same column.
            # This prime must exist
            for l in range(m):
                if Primes[k, l] > 0:
                    break
            # print('Found prime at (%d, %d) in step 4' % (k, l))
            # un-prime it and add it to the path
            Primes[k, l] = 0
            path[T, 0] = k
            path[T, 1] = l
            T += 1
        # star all locations on the path
        for t in range(T):
            Stars[path[t, 0], path[t, 1]] = 1
        # remove all primes
        Primes[:, :] = 0
        # update all column covers
        col_covers[:] = 0
        for j in range(m):
            for i in range(m):
                if Stars[i, j] > 0:
                    col_covers[j] = 1
                    break
        # step 4 complete, return to step 2
        # print('new stars after step 4\n%s' % str(np.asarray(Stars)))
        # print('new col covers after step 4: %s' % str(np.asarray(col_covers)))

def munkres(C):
    """ This calls the Munkres algorithm to find a minimal
    matching for the given cost matrix C. Note that this function
    is more for debugging purposes. If you want to call this
    algorithm from Python, you are better served by calling
    scipy.optimize.linear_sum_assignment.

    Parameters
    ----------
    C: ndarray
        An m x m cost matrix.

    Returns
    -------
    pi: ndarray
        An m-element array where pi[i] is the index to which i is
        assigned.

    """
    # set up all temporary variables
    C = np.copy(C)
    m = C.shape[0]
    Stars  = np.zeros((m, m), dtype=np.intc)
    Primes = np.zeros((m, m), dtype=np.intc)
    row_covers = np.zeros(m, dtype=np.intc)
    col_covers = np.zeros(m, dtype=np.intc)
    path = np.zeros((m, 2), dtype=int)
    pi = np.zeros(m, dtype=int)
    munkres_(C, Stars, Primes, row_covers, col_covers, path, pi)
    return pi



#########################
# Backtracing Functions #
#########################

cdef double _BACKTRACE_TOL = 1E-5

def uted_backtrace(x_nodes, x_adj, y_nodes = None, y_adj = None, delta = None):
    """ Computes the unordered tree edit distance between the trees x and y,
    each described by a list of nodes and an adjacency list adj, where adj[i]
    is a list of indices pointing to children of node i. This function
    returns an alignment representation of the distance.

    Note that we assume a proper depth-first-search order of adj, i.e. for
    every node i, the following indices are all part of the subtree rooted at
    i until we hit the index of i's right sibling or the end of the tree.

    Parameters
    ----------
    x_nodes: list or tuple
        a list of nodes for tree x OR a tuple of the form (x_nodes, x_adj).
    x_adj: list or tuple
        an adjacency list for tree x OR a tuple of the form (y_nodes, y_adj).
    y_nodes: list (default = x_adj[0])
        a list of nodes for tree y.
    y_adj: list (default = x_adj[1])
        an adjacency list for tree y.
    delta: function (default = None)
        a function that takes two nodes as inputs and returns their pairwise
        distance, where delta(x, None) should be the cost of deleting x and
        delta(None, y) should be the cost of inserting y. If undefined, this
        method calls standard_ted instead.

    Returns
    -------
    alignment: class alignment.Alignment
        A co-optimal alignment between x and y according to the unordered tree
        edit distance.

    """
    if isinstance(x_nodes, tuple):
        x_nodes, x_adj, y_nodes, y_adj = extract_from_tuple_input(x_nodes, x_adj)

    # the number of nodes in both trees
    cdef int m = len(x_nodes)
    cdef int n = len(y_nodes)

    # if either tree is empty, we can only delete/insert all nodes in the
    # non-empty tree.
    cdef int i = 0
    cdef int j = 0
    ali = Alignment()
    if m == 0:
        for j in range(n):
            ali.append_tuple(-1, j)
        return ali
    if n == 0:
        for i in range(n):
            ali.append_tuple(i, -1)
        return ali

    # Set up an array to store edit costs for replacements,
    # deletions, and insertions
    Delta = np.ones((m+1, n+1))
    cdef double[:,:] Delta_view = Delta

    if delta is None:
        for i in range(m):
            for j in range(n):
                if x_nodes[i] == y_nodes[j]:
                    Delta_view[i, j] = 0.
    else:
        # First, compute all pairwise replacement costs
        for i in range(m):
            for j in range(n):
                Delta_view[i,j] = delta(x_nodes[i], y_nodes[j])
        # Then, compute the deletion and insertion costs
        for i in range(m):
            Delta_view[i,n] = delta(x_nodes[i], None)
        for j in range(n):
            Delta_view[m,j] = delta(None, y_nodes[j])

    # compute the actual tree edit distance
    D_forest, D_tree = _uted(x_nodes, x_adj, y_nodes, y_adj, Delta)

    # set up temporary variables for backtracing
    cdef int k
    cdef int l
    cdef int i_k
    cdef int j_l
    cdef int m_i
    cdef int n_j
    cdef double rep_cost = 0.
    cdef double del_cost = 0.
    cdef int k_min = 0
    cdef double ins_cost = 0.
    cdef int l_min = 0

    # start backtracing. We perform backtracing via a stack which
    # stores a tuple of indices (i, j) in both trees and a 'mode'
    # flag, namely whether we are currently align the subtrees at
    # (i, j) or the child forests.
    stk = [(0, 0, 0)]
    while stk:
        i, j, mode = stk.pop()
        # print('pop (%d, %d, %d)' % (i, j, mode))
        if mode == 0:
            # first, check if we only wish to delete/insert
            if i == -1:
                # print('insert %d' % j)
                ali.append_tuple(-1, j)
                for j_l in y_adj[j][::-1]:
                    stk.append((-1, j_l, 0))
                continue
            if j == -1:
                # print('delete %d' % i)
                ali.append_tuple(i, -1)
                for i_k in x_adj[i][::-1]:
                    stk.append((i_k, -1, 0))
                continue
            # print('continuing tree alignment at (%d, %d)' % (i, j))
            m_i = len(x_adj[i])
            n_j = len(y_adj[j])
            # check whether replacement is co-optimal
            # print('D_tree[i, j] = %g; Delta[i, j] = %g, D_forest[i, j] = %g' % (D_tree[i, j], Delta[i, j], D_forest[i, j]))
            if D_tree[i, j] + _BACKTRACE_TOL > Delta[i, j] + D_forest[i, j]:
                # print('replacement co-optimal')
                # if so, i and j are aligned
                ali.append_tuple(i, j)
                # and we need to optimally align all children
                stk.append((i, j, 1))
                continue
            # check whether deletion is co-optimal
            if m_i == 0:
                if D_tree[i, j] + _BACKTRACE_TOL > Delta[i, n] + D_tree[m, j]:
                    # print('deletion co-optimal')
                    # if it is and i has no children,
                    # delete i and insert everything in j
                    ali.append_tuple(i, -1)
                    stk.append((-1, j, 0))
                    continue
            else:
                # we consider all options to delete i and replace
                # it with one of its children
                for k in range(len(x_adj[i])):
                    i_k = x_adj[i][k]
                    del_cost = D_tree[i_k, j] + D_tree[i, n] - D_tree[i_k, n]
                    if D_tree[i, j] + _BACKTRACE_TOL > del_cost:
                        k_min = k
                        break
                if D_tree[i, j] + _BACKTRACE_TOL > del_cost:
                    # print('deletion with child %d co-optimal' % k_min)
                    # delete i
                    ali.append_tuple(i, -1)
                    for k in range(len(x_adj[i])-1, -1, -1):
                        i_k = x_adj[i][k]
                        if k == k_min:
                            # continue the alignment between i_k and j
                            stk.append((i_k, j, 0))
                        else:
                            # delete all children of i except i_k
                            stk.append((i_k, -1, 0))
                    continue
            # check whether insertion is co-optimal
            if n_j == 0:
                if D_tree[i, j] + _BACKTRACE_TOL > Delta[m, j] + D_tree[i, n]:
                    # print('insertion co-optimal')
                    # if it is and j has no children,
                    # insert j and delete everything in i
                    ali.append_tuple(-1, j)
                    stk.append((i, -1, 0))
                    continue
            else:
                # we consider all options to insert j and replace
                # it with one of its children
                for l in range(len(y_adj[j])):
                    j_l = y_adj[j][l]
                    ins_cost = D_tree[i, j_l] + D_tree[m, j] - D_tree[m, j_l]
                    if D_tree[i, j] + _BACKTRACE_TOL > ins_cost:
                        l_min = l
                        break
                if D_tree[i, j] + _BACKTRACE_TOL > ins_cost:
                    # print('insertion with child %d co-optimal' % l_min)
                    # insert j
                    ali.append_tuple(-1, j)
                    for l in range(len(y_adj[j])-1, -1, -1):
                        j_l = y_adj[j][l]
                        if l == l_min:
                            # continue the alignment between i and j_l
                            stk.append((i, j_l, 0))
                        else:
                            # delete all children of j except j_l
                            stk.append((-1, j_l, 0))
                    continue
        elif mode == 1:
            # print('continuing forest alignment at (%d, %d)' % (i, j))
            m_i = len(x_adj[i])
            n_j = len(y_adj[j])
            # if the children are empty, we need to delete all
            # children of the counterpart
            if m_i == 0:
                for j_l in y_adj[j][::-1]:
                    stk.append((-1, j_l, 0))
                continue
            if n_j == 0:
                for i_k in x_adj[i][::-1]:
                    stk.append((i_k, -1, 0))
                continue
            # next, check if replacement is co-optimal. For this
            # purpose, perform the Munkres/Hungarian algorithm
            # for an optimal alignment

            # prepare a cost matrix for the algorithm
            C = np.zeros((m_i + n_j, m_i + n_j))
            for k in range(m_i):
                i_k = x_adj[i][k]
                for l in range(n_j):
                    j_l = y_adj[j][l]
                    # matching ci with cj means editing the ci'th
                    # child of i to the cj'th child of j.
                    C[k, l] = D_tree[i_k, j_l]
            C[:m_i, n_j:] = INFINITY
            for k in range(m_i):
                # matching c with n_j + c means deleting the
                # c'th child of i
                i_k = x_adj[i][k]
                C[k, n_j + k] = D_tree[i_k, n]
            C[m_i:, :n_j] = INFINITY
            for l in range(n_j):
                # matching m_i + c with c means inserting the
                # c'th child of j
                j_l = y_adj[j][l]
                C[m_i + l, l] = D_tree[m, j_l]
            C[m_i:, n_j:] = 0.
            # call the munkres algorithm
            pi = munkres(C)
            # compute the cost
            rep_cost = 0.
            for k in range(m_i):
                i_k = x_adj[i][k]
                if pi[k] >= n_j:
                    # if pi[k] >= n_j, tree i_k should be deleted
                    rep_cost += D_tree[i_k, n]
                else:
                    # otherwise we replace i_k with j_pi[k]
                    j_l = y_adj[j][pi[k]]
                    rep_cost += D_tree[i_k, j_l]
            for l in range(n_j):
                j_l = y_adj[j][l]
                if pi[m_i + l] < n_j:
                    # if pi[m_i + l] < n_j, tree j_l should be inserted
                    rep_cost += D_tree[m, j_l]
            # check if forest replacement is optimal
            if D_forest[i, j] + _BACKTRACE_TOL > rep_cost:
                # print('forest replacement with alignment %s co-optimal' % str(pi))
                # if replacement is co-optimal, perform all
                # deletion/insertion/replacements of subtrees
                # dictated by the Hungarian result
                for l in range(n_j-1, -1, -1):
                    if pi[m_i + l] < n_j:
                        # insert everything in subtree j_l
                        stk.append((-1, y_adj[j][l], 0))
                for k in range(m_i-1, -1, -1):
                    if pi[k] < n_j:
                        # put subtree replacements onto the stack
                        stk.append((x_adj[i][k], y_adj[j][pi[k]], 0))
                    else:
                        # delete everything in subtree i_k
                        stk.append((x_adj[i][k], -1, 0))
                continue
            # if it is not, check if forest deletion is optimal
            for k in range(m_i):
                i_k = x_adj[i][k]
                # accordingly, we need to consider the cost of editing
                # the children of node i_k with the children of j,
                # plus the cost of deleting all other children of i.
                del_cost = D_forest[i_k, j] + D_forest[i, n] - D_forest[i_k, n]
                if D_forest[i, j] + _BACKTRACE_TOL > del_cost:
                    k_min = k
                    break
            if D_forest[i, j] + _BACKTRACE_TOL > del_cost:
                # print('forest deletion with child %d co-optimal' % k_min)
                for k in range(m_i):
                    i_k = x_adj[i][k]
                    if k == k_min:
                        # delete i_k
                        ali.append_tuple(i_k, -1)
                        # continue the forest alignment between
                        # the children of i_k and j
                        stk.append((i_k, j, 1))
                    else:
                        # delete all children of i except i_k
                        stk.append((i_k, -1, 0))
                continue
            # if it is not, check if forest insertion is optimal
            for l in range(n_j):
                j_l = y_adj[j][l]
                # accordingly, we need to consider the cost of editing
                # tree i into tree j_l plus the cost o inserting all other
                # children of j.
                ins_cost = D_tree[i, j_l] + D_tree[m, j] - D_tree[m, j_l]
                if D_forest[i, j] + _BACKTRACE_TOL > ins_cost:
                    l_min = l
                    break
            if D_forest[i, j] + _BACKTRACE_TOL > ins_cost:
                # print('forest insertion with child %d co-optimal' % l_min)
                for l in range(n_j):
                    j_l = y_adj[j][l]
                    if l == l_min:
                        # insert j_l
                        ali.append_tuple(-1, j_l)
                        # continue the forest alignment between
                        # the children of i and j_l
                        stk.append((i, j_l, 1))
                    else:
                        # insert all children of j except j_l
                        stk.append((-1, j_l, 0))
            continue
        else:
            raise ValueError('Internal error: Unrecognized mode: %s' % str(mode))
        # if nothing is co-optimal, raise an exception
        raise ValueError('Internal error: No option was co-optimal during alignment of (%d, %d), mode = %s' % (i, j, str(mode)))
    return ali



import setuptools
from setuptools.extension import Extension

try:
    # try to import CYTHON
    from Cython.Build import cythonize
    USE_CYTHON = True
except ImportError:
    # if it fails, we use c compilation
    USE_CYTHON = False

ext = '.pyx' if USE_CYTHON else '.c'

extensions = [
    Extension(
        "edist.adp",
        ["edist/adp" + ext],
    ),
    Extension(
        "edist.dtw",
        ["edist/dtw" + ext],
    ),
    Extension(
        "edist.sed",
        ["edist/sed" + ext],
    ),
    Extension(
        "edist.ted",
        ["edist/ted" + ext],
    ),
    Extension(
        "edist.uted",
        ["edist/uted" + ext],
    ),
    Extension(
        "edist.seted",
        ["edist/seted" + ext],
    ),
]

if(USE_CYTHON):
    extensions = cythonize(extensions)

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="edist",
    version="1.2.0",
    author="Benjamin Paassen",
    author_email="bpaassen@techfak.uni-bielefeld.de",
    description="Edit distance implementations in cython",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.ub.uni-bielefeld.de/bpaassen/python-edit-distances",
    packages=['edist'],
    install_requires=['numpy', 'scikit-learn', 'scipy', 'proto-dist-ml'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    keywords='levenshtein-distance dynamic-time-warping sequence-edit-distance sequence-alignment tree-edit-distance algebraic-dynamic-programming',
    ext_modules=extensions
)
